from __future__ import annotations

import logging

from nabu.stitching.config import StitchingType
from nabu.stitching.config import identifiers_as_str_to_instances

from nxtomomill.models.utils import convert_str_to_tuple as _convert_str_to_tuple
from silx.gui import qt
from tomoscan.series import Series
from tomoscan.scanbase import TomoScanBase as _TomoScanBase
from tomoscan.volumebase import VolumeBase as _VolumeBase

from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.volume.volumefactory import VolumeFactory
from tomwer.core.volume.volumebase import TomwerVolumeBase
from tomwer.core.tomwer_object import TomwerObject
from tomwer.gui.stitching.stitching_preview import PreviewStitchingPlot
from tomwer.gui.stitching.stitching_raw import RawStitchingPlot
from tomwer.gui.stitching.axisorderedlist import EditableOrderedTomoObjWidget
from .singleaxis import _SingleAxisMixIn

_logger = logging.getLogger(__name__)


def convert_str_to_tuple(input_str, none_if_empty):
    if input_str is None:
        return None
    elif isinstance(input_str, (tuple, list)):
        return tuple(input_str)
    else:
        return _convert_str_to_tuple(input_str=input_str, none_if_empty=none_if_empty)


class SingleAxisStitchingWidget(qt.QWidget, _SingleAxisMixIn):
    """
    Main widget for (single axis) stitching. It contains display and options definition.
    """

    sigStitchingTypeChanged = qt.Signal(str)
    """emit when stitching type changes"""
    sigTomoObjsLoaded = qt.Signal(tuple)
    """Signal emit when during setting a configuration this trigger some addition of tomo object"""

    def __init__(self, axis: int, parent=None) -> None:
        super().__init__(parent)
        assert axis in (0, 1, 2)
        self._axis = axis  # used from _SingleAxisMixIn interface
        self.setLayout(qt.QGridLayout())

        self._stitchingTypeCB = qt.QComboBox(parent=self)

        # filter stitching types that does belong to the axis (GUI is specific to a given axis. The other have no sense at this level)
        stitching_types = filter(
            lambda s_type: s_type.lower().startswith(self.axis_alias(axis).lower()),
            [item.value for item in StitchingType],
        )
        for mode in stitching_types:
            self._stitchingTypeCB.addItem(mode)
        self._stitchingTypeCB.currentIndexChanged.connect(self._stitchingTypeChanged)
        self.layout().addWidget(qt.QLabel("stitching method:"), 0, 0, 1, 1)
        self.layout().addWidget(self._stitchingTypeCB, 0, 1, 1, 1)

        self._mainWidget = _SingleAxisStitchingCentralTabWidget(parent=self, axis=axis)
        self.layout().addWidget(self._mainWidget, 1, 0, 4, 4)

        # set up
        self.setStitchingType(self.getStitchingType())
        self._mainWidget.setCurrentWidget(self._mainWidget._previewPlot)

        # connect signal / slot
        self._stitchingTypeCB.currentIndexChanged.connect(self._stitchingTypeChanged)

    def close(self):
        self._mainWidget.close()
        # requested for the waiting plot update
        super().close()

    def clean(self):
        self._mainWidget.clean()

    def _stitchingTypeChanged(self, *args, **kwargs):
        self.sigStitchingTypeChanged.emit(self.getStitchingType().value)

    def getStitchingType(self):
        return StitchingType(self._stitchingTypeCB.currentText())

    def setStitchingType(self, mode):
        mode = StitchingType(mode)
        idx = self._stitchingTypeCB.findText(mode.value)
        if idx >= 0:
            self._stitchingTypeCB.setCurrentIndex(idx)

    def get_available_pre_processing_stitching_mode(self):
        pre_proc_modes: tuple[str] = tuple(
            filter(
                lambda mode: "preproc" in mode,
                [
                    self._stitchingTypeCB.itemText(i_item)
                    for i_item in range(self._stitchingTypeCB.count())
                ],
            )
        )
        return tuple([StitchingType(mode) for mode in pre_proc_modes])

    def get_available_post_processing_stitching_mode(self):
        post_proc_modes: tuple[str] = tuple(
            filter(
                lambda mode: "postproc" in mode,
                [
                    self._stitchingTypeCB.itemText(i_item)
                    for i_item in range(self._stitchingTypeCB.count())
                ],
            )
        )
        return tuple([StitchingType(mode) for mode in post_proc_modes])

    def addTomoObj(self, tomo_obj: TomwerObject):
        self._mainWidget.addTomoObj(tomo_obj)
        self._updatePreviewPixelSize()

    def getTomoObjs(self) -> tuple:
        return self._mainWidget.getTomoObjs()

    def removeTomoObj(self, tomo_obj: TomwerObject):
        self._mainWidget.removeTomoObj(tomo_obj)
        self._updatePreviewPixelSize()

    def _updatePreviewPixelSize(self):
        """update the pixel size of the preview from existing tomo obj"""

        def get_pixel_size():
            tomo_objs = self._mainWidget.getTomoObjs()
            for tomo_obj in tomo_objs:
                if (
                    isinstance(tomo_obj, NXtomoScan)
                    and tomo_obj.sample_x_pixel_size is not None
                    and tomo_obj.sample_y_pixel_size is not None
                ):
                    return tomo_obj.sample_x_pixel_size, tomo_obj.sample_y_pixel_size
                elif (
                    isinstance(tomo_obj, TomwerVolumeBase)
                    and tomo_obj.voxel_size is not None
                ):
                    # warning: axis are inverted between the volume and the voxel_size. voxel_size are given as (x, y, z) instead of (z, y, x)
                    return (
                        tomo_obj.voxel_size[2 - self.first_axis],
                        tomo_obj.voxel_size[2 - self.first_axis],
                    )
            return None, None

        pixel_size = get_pixel_size()
        self._mainWidget._previewPlot.setPixelSize(pixel_size_m=pixel_size)

    def getConfiguration(self) -> dict:
        # missing parameters:
        # * overwrite
        # * slices
        # * slurm stuff...

        tomo_objs = self._mainWidget.getTomoObjs()

        def filter_empty_list_and_cast_as_int(elmts):
            new_list = [int(elmt) for elmt in elmts if elmt is not None]
            if len(new_list) == 0:
                return None
            else:
                return elmts

        first_axis_pos_px = filter_empty_list_and_cast_as_int(
            [
                obj.stitching_metadata.get_abs_position_px(axis=self.first_axis) or 0
                for obj in tomo_objs
            ]
        )
        second_axis_pos_px = filter_empty_list_and_cast_as_int(
            [
                obj.stitching_metadata.get_abs_position_px(axis=self.second_axis) or 0
                for obj in tomo_objs
            ]
        )
        return {
            "stitching": {
                "type": self.getStitchingType().value,
                f"axis_{self.first_axis}_pos_px": (
                    "" if first_axis_pos_px is None else first_axis_pos_px
                ),
                f"axis_{self.second_axis}_pos_px": (
                    "" if second_axis_pos_px is None else second_axis_pos_px
                ),
            },
            "inputs": {
                "input_datasets": [obj.get_identifier().to_str() for obj in tomo_objs],
            },
        }

    def setConfiguration(self, config: dict) -> None:
        stitching_type = config.get("stitching", {}).get("type", None)
        if stitching_type is not None:
            self.setStitchingType(stitching_type)
        tomo_obj_ids = config.get("inputs", {}).get("input_datasets", None)
        tomo_obj_ids = identifiers_as_str_to_instances(tomo_obj_ids)
        first_axis_pos = convert_str_to_tuple(
            config.get("stitching", {}).get(f"axis_{self.first_axis}_pos_px", None),
            none_if_empty=True,
        )
        second_axis_pos = convert_str_to_tuple(
            config.get("stitching", {}).get(f"axis_{self.second_axis}_pos_px", None),
            none_if_empty=True,
        )
        if tomo_obj_ids is not None:
            self._mainWidget.clearTomoObjs()
            if first_axis_pos is None:
                first_axis_pos = [None] * len(tomo_obj_ids)
            if second_axis_pos is None:
                second_axis_pos = [None] * len(tomo_obj_ids)
            if len(first_axis_pos) != len(tomo_obj_ids):
                _logger.error(
                    f"incoherent axis {self.first_axis} position compared to the number of input datasets. Will ignore those"
                )
                first_axis_pos = [None] * len(tomo_obj_ids)
            if len(second_axis_pos) != len(tomo_obj_ids):
                _logger.error(
                    f"incoherent axis {self.second_axis} position compared to the number of input datasets. Will ignore those"
                )
                second_axis_pos = [None] * len(tomo_obj_ids)

            new_tomo_objs = []
            for tomo_obj_id, first_axis_v, second_axis_v in zip(
                tomo_obj_ids, first_axis_pos, second_axis_pos
            ):
                if isinstance(tomo_obj_id, TomwerObject):
                    tomo_obj = tomo_obj_id
                elif isinstance(tomo_obj_id, _TomoScanBase):
                    # for now we need to convert it back because object are not the same
                    tomo_obj = ScanFactory.create_tomo_object_from_identifier(
                        tomo_obj_id.get_identifier().to_str()
                    )
                elif isinstance(tomo_obj_id, _VolumeBase):
                    tomo_obj = VolumeFactory.create_tomo_object_from_identifier(
                        tomo_obj_id.get_identifier().to_str()
                    )
                else:
                    tomo_obj = ScanFactory.create_tomo_object_from_identifier(
                        tomo_obj_id
                    )
                self.addTomoObj(tomo_obj=tomo_obj)
                # set metadata information if any
                for axis, axis_value in zip((0, 1), (first_axis_v, second_axis_v)):
                    if axis_value is not None:
                        tomo_obj.stitching_metadata.setPxPos(int(axis_value), axis=axis)
                new_tomo_objs.append(tomo_obj)
            self.sigTomoObjsLoaded.emit(tuple(new_tomo_objs))
            self._mainWidget._axisOrderedList._orderMightHaveChanged()

    # expose API
    def setAddTomoObjCallbacks(self, *args, **kwargs):
        self._mainWidget.setAddTomoObjCallbacks(*args, **kwargs)

    def setRemoveTomoObjCallbacks(self, *args, **kwargs):
        self._mainWidget.setRemoveTomoObjCallbacks(*args, **kwargs)


class _SingleAxisStitchingCentralTabWidget(qt.QTabWidget):
    """
    Tab widget containing:
    * preview interface of the stitching
    * ordered list of the scan / volume position over the main stitching axis
    """

    def __init__(self, axis: int, parent=None) -> None:
        super().__init__(parent)
        assert axis in (0, 1, 2)
        self._seriesName = None
        self._axisOrderedList = EditableOrderedTomoObjWidget(parent=self, axis=axis)
        self.addTab(self._axisOrderedList, f"axis {axis} ordered list")
        self._previewPlot = PreviewStitchingPlot(parent=self, axis=axis)
        self.addTab(self._previewPlot, "stitching preview")
        # TODO: add a raw display to print frame from raw position z positions ...
        self._rawDisplayPlot = RawStitchingPlot(
            parent=self,
            aspectRatio=True,
            logScale=False,
            copy=False,
            save=False,
            print_=False,
            grid=False,
            curveStyle=False,
            mask=False,
            alpha_values=True,
        )
        self._rawDisplayPlot.setKeepDataAspectRatio(True)
        self._rawDisplayPlot.setAxesDisplayed(False)
        self.addTab(self._rawDisplayPlot, "raw display")
        # add an option to activate / deactivate auto update of the raw display as it can be time consuming.
        raw_display_idx = self.indexOf(self._rawDisplayPlot)
        self._rawDisplayCB = qt.QCheckBox(self)
        self.tabBar().setTabButton(
            raw_display_idx,
            qt.QTabBar.LeftSide,
            self._rawDisplayCB,
        )
        self.setTabToolTip(
            raw_display_idx,
            "If toggled will keep the raw display up to date from axis 0 modifications",
        )
        # set up: turn overlay one by default
        self._previewPlot._backGroundAction.setChecked(True)

    def _handleRawDisplayConnection(self, toggled: bool):
        if toggled:
            self._connectRawDisplayConnection()
        else:
            self._disconnectRawDisplayConnection()

    def setSeries(self, series: Series):
        for elmt in series:
            self._axisOrderedList.addTomoObj(elmt)
        self.setSeriesName(series.name)

    def addTomoObj(self, tomo_obj: TomwerObject):
        self._axisOrderedList.addTomoObj(tomo_obj)

    def removeTomoObj(self, tomo_obj: TomwerObject):
        self._axisOrderedList.removeTomoObj(tomo_obj=tomo_obj)

    def getSeriesName(self) -> str:
        return self._seriesName

    def setSeriesName(self, name: str):
        self._seriesName = name

    def getTomoObjs(self) -> tuple:
        return self._axisOrderedList.getTomoObjsAxisOrdered()

    def clearTomoObjs(self):
        self._axisOrderedList.clearTomoObjs()

    def clean(self) -> None:
        self.clearTomoObjs()
        self._previewPlot.clear()

    def close(self):
        self._previewPlot.close()
        # requested for the waiting plot update
        super().close()

    def setAddTomoObjCallbacks(self, *args, **kwargs):
        self._axisOrderedList.setAddTomoObjCallbacks(*args, **kwargs)

    def setRemoveTomoObjCallbacks(self, *args, **kwargs):
        self._axisOrderedList.setRemoveTomoObjCallbacks(*args, **kwargs)



"""
Contains tools used by the orange add-on and used to keep the orange-add-on and
gui 'alive', avoid gui freeze during processing.

With time it should be replace by the ewoksorange mechanism (inheritance to ewoksorange widgets)
"""



# coding: utf-8
from __future__ import annotations


import os
import pytest

import numpy

from tomwer.core.process.reconstruction.saaxis.params import SAAxisParams
from tomwer.core.process.reconstruction.saaxis.saaxis import SAAxisTask
from tomwer.core.process.reconstruction.scores.scores import (
    _METHOD_TO_FCT,
    compute_score_contrast_std,
)
from tomwer.core.utils.scanutils import MockNXtomo


def test_img_contrast_std_score():
    """simple test of the API to call compute_score_contrast_std"""
    data = numpy.random.random(100 * 100).reshape(100, 100)
    compute_score_contrast_std(data)


@pytest.mark.parametrize("fct", _METHOD_TO_FCT.values())
def test_method_to_function(fct):
    """Test the dictionary used to for linking the score method to the
    callback function"""
    data = numpy.random.random(100 * 100).reshape(100, 100)
    res = fct(data)
    assert res is not None
    assert isinstance(res, float)


def testSAAxisProcess(tmp_path):
    """Test the SAAxisProcess class"""
    scan_path = tmp_path / "mock_nxtomo"
    scan_path.mkdir()
    # set up
    dim = 10
    mock = MockNXtomo(
        scan_path=scan_path, n_proj=10, n_ini_proj=10, scan_range=180, dim=dim
    )
    mock.add_alignment_radio(index=10, angle=90)
    mock.add_alignment_radio(index=10, angle=0)
    scan = mock.scan

    default_saaxis_params = SAAxisParams()
    default_saaxis_params.output_dir = os.path.join(scan_path, "output_dir")
    default_saaxis_params.slice_indexes = {"slice": 4}
    default_saaxis_params.nabu_config = {}
    default_saaxis_params.dry_run = True
    default_saaxis_params.file_format = "hdf5"

    # test processing
    process = SAAxisTask(
        inputs={
            "data": scan,
            "sa_axis_params": default_saaxis_params.to_dict(),
            "serialize_output_data": False,
            "dry_run": True,
        }
    )

    default_saaxis_params.estimated_cor = 11
    default_saaxis_params.research_width = 2
    process = SAAxisTask(
        inputs={
            "data": scan,
            "sa_axis_params": default_saaxis_params.to_dict(),
            "serialize_output_data": False,
            "dry_run": True,
        },
    )
    process.run()
    process = SAAxisTask(
        inputs={
            "data": scan,
            "sa_axis_params": default_saaxis_params.to_dict(),
            "serialize_output_data": False,
            "dry_run": True,
        },
    )
    process.run()



# coding: utf-8
from __future__ import annotations


import gc

from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt

from orangecontrib.tomwer.widgets.control.TimerOW import TimerOW


class TestTimerOW(TestCaseQt):
    def setUp(self):
        super().setUp()
        self.widget = TimerOW()

    def tearDown(self):
        self.widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.widget.close()
        self.widget = None
        gc.collect()

    def test(self):
        self.widget.show()



# coding: utf-8
from __future__ import annotations


import logging
import socket

from ._QThreadListenerBase import QThreadListenerBase

_logger = logging.getLogger(__name__)

try:
    from blissdata.redis_engine.store import DataStore
    from blissdata.redis_engine.scan import ScanState
    from blissdata.beacon.data import BeaconData

    try:
        from blissdata.exceptions import NoScanAvailable
    except ImportError:
        from blissdata.redis_engine.exceptions import NoScanAvailable
except ImportError:
    _logger.warning("blissdata not installed. Cannot run the Blissdata Listener.")
    has_blissdata = False
else:
    has_blissdata = True


_logger = logging.getLogger(__name__)


class BlissDataListenerQThread(QThreadListenerBase):
    """Implementation of the _BaseDataListenerThread with a QThread.

    On this use case we use directly blissdata (requires Bliss version 2.0 or higher)

    # note: on bliss-tomo:
    # a sequence starts that triggers other scans.
    # then those scans starts and end...
    # and then only the parent sequence ends.
    # so the listener must be started before the sequence starts... which is already the case.

    ..warning:: this thread will never trigger 'sigScanAdded' or 'sigServerStop' as the first one is indeed never used at the moment and the second one is a specific feature for the json-rpc mechanism
    """

    def __init__(self, host, port, acquisitions):
        super().__init__(host=host, port=port, acquisitions=acquisitions)
        self._host = host
        self._port = port
        self._stop = False

        if not has_blissdata:
            raise RuntimeError("'blissdata' not installed. Cannot create the listener")

        try:
            self.createDataStore()
        except socket.gaierror as e:
            _logger.error(
                f"Fail to connect to host: {self._host}:{self._port}",
                exc_info=e,
                stack_info=True,
                stacklevel=3,
            )
            self._canConnect = False
        else:
            self._canConnect = True

    def createDataStore(self):
        self._beacon_client = BeaconData(host=self._host, port=self._port)
        self._redis_url = self._beacon_client.get_redis_data_db()
        self._data_store = DataStore(self._redis_url)

    def run(self):
        if not self._canConnect:
            return

        while not self._stop:
            try:
                _, key = self._data_store.get_next_scan(timeout=1)
            except (TimeoutError, NoScanAvailable):
                # not: Today the `NoScanAvailable` will be raised in case of a time out.
                continue
            except Exception as e:
                raise Exception("Failed to access next scan.") from e

            scan_sequence = self._data_store.load_scan(key)
            if not self.isBlissTomoScan(scan_sequence=scan_sequence):
                _logger.warning(
                    f"Scan {scan_sequence} is not recognized as a bliss-tomo acquisition. It will be ignored."
                )
                continue

            saving_file = scan_sequence.info["filename"]
            entry = scan_sequence.info["scan_nb"]
            proposal_file = scan_sequence.info["nexuswriter"]["masterfiles"]["proposal"]
            sample_file = scan_sequence.info["nexuswriter"]["masterfiles"][
                "dataset_collection"
            ]

            self.sigAcquisitionStarted.emit(
                (saving_file, entry, proposal_file, sample_file)
            )

            while scan_sequence.state < ScanState.CLOSED:
                if self._stop:
                    return
                scan_sequence.update()

            succeeded = scan_sequence.info["end_reason"] == "SUCCESS"
            self.sigAcquisitionEnded.emit(
                (saving_file, entry, proposal_file, sample_file, succeeded)
            )

    @staticmethod
    def isBlissTomoScan(scan_sequence) -> bool:
        """Check if the scan sequence is a bliss-tomo sequence"""
        tomo_config_group = scan_sequence.info.get("technique", {}).get(
            "tomoconfig", None
        )
        return tomo_config_group is not None

    def stop(self):
        self._stop = True
        self._beacon_client = None
        self._redis_url = None
        self._data_store = None



# coding: utf-8

"""
contains gui relative to semi-automatic axis calculation
"""

from __future__ import annotations


import logging
from typing import Iterable

import numpy
from silx.gui import qt

from tomwer.core.process.reconstruction.nabu.utils import (
    _NabuMode,
    retrieve_lst_of_value_from_str,
)
from tomwer.gui.configuration.level import ConfigurationLevel
from tomwer.core.process.reconstruction.sadeltabeta.sadeltabeta import (
    SADeltaBetaTask,
)
from tomwer.core.process.reconstruction.scores.params import ScoreMethod
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.gui import icons
from tomwer.gui.reconstruction.nabu.nabuconfig.phase import _NabuPhaseConfig
from tomwer.gui.reconstruction.nabu.platform import NabuPlatformSettings
from tomwer.gui.reconstruction.saaxis.sliceselector import SliceSelector
from tomwer.gui.reconstruction.scores.control import ControlWidget
from tomwer.gui.reconstruction.scores.scoreplot import DelaBetaSelection
from tomwer.gui.reconstruction.scores.scoreplot import ScorePlot as _ScorePlot
from tomwer.gui.utils.buttons import TabBrowsersButtons
from tomwer.gui.utils.scandescription import ScanNameLabelAndShape
from tomwer.gui.settings import TAB_LABEL_PLATFORM_SETTINGS
from tomwer.synctools.sadeltabeta import QSADeltaBetaParams
from tomwer.gui.reconstruction.sacommon import NabuWidgetWithToolbar as NabuWidget

_logger = logging.getLogger(__name__)


class ScorePlot(_ScorePlot, constructor=DelaBetaSelection):
    """Score plot dedicated to center delta / beta values."""

    def _updateScores(self):
        scan = self.__scan() if self.__scan else None
        if scan is not None:
            assert isinstance(scan, TomwerScanBase)
            if scan.sa_delta_beta_params:
                scan.sa_delta_beta_params.score_method = self.getScoreMethod()
                # update autofocus
                SADeltaBetaTask.autofocus(scan)
        self.setVarScores(
            scores=self._scores,
            score_method=self.getScoreMethod(),
            update_only_scores=True,
        )

    def _applyAutofocus(self):
        scan = self.__scan() if self.__scan else None
        if scan is None:
            return
        if scan.sa_delta_beta_params:
            best_db = scan.sa_delta_beta_params.autofocus
            if best_db:
                self._varSlider.setVarValue(best_db)


class _SADeltaBetaTabWidget(qt.QTabWidget):
    sigConfigurationChanged = qt.Signal()
    """Signal emit when the configuration changes"""

    def __init__(self, parent=None, backend=None):
        qt.QTabWidget.__init__(self, parent=parent)

        self._deltaBetaSelectionWidget = DeltaBetaSelectionWidget(self)
        delta_beta_icon = icons.getQIcon("delta_beta")
        self.addTab(
            self._deltaBetaSelectionWidget, delta_beta_icon, "delta beta values"
        )
        # nabu reconstruction parameters
        self._nabuSettings = NabuWidget(self)
        self._nabuSettings.setConfigurationLevel(level="required")
        self._nabuSettings.hidePaganinInterface()
        self._nabuSettings.hideSlicesInterface()
        nabu_icon = icons.getQIcon("nabu")
        self.addTab(self._nabuSettings, nabu_icon, "reconstruction settings")
        # platform settings
        self._localPlatformSettings = NabuPlatformSettings(self)
        settings_icons = icons.getQIcon("parameters")
        self.addTab(
            self._localPlatformSettings, settings_icons, TAB_LABEL_PLATFORM_SETTINGS
        )
        # results
        self._resultsViewer = ScorePlot(self, variable_name="db", backend=backend)
        results_icon = icons.getQIcon("results")
        self.addTab(self._resultsViewer, results_icon, "reconstructed slices")

        # connect signal / slot
        self._nabuSettings.sigConfigChanged.connect(self._configurationChanged)
        self._deltaBetaSelectionWidget.sigConfigurationChanged.connect(
            self._configurationChanged
        )
        self._resultsViewer.sigConfigurationChanged.connect(self._configurationChanged)

        # expose function API
        self.setNabuReconsParams = self._nabuSettings.setConfiguration
        self.getNabuReconsParams = self._nabuSettings.getConfiguration
        self.saveReconstructedSlicesTo = self._resultsViewer.saveReconstructedSlicesTo

    def setDeltaBetaScores(self, *args, **kwargs):
        self._resultsViewer.setVarScores(*args, **kwargs)

    def setCurrentVarValue(self, *args, **kwargs):
        self._resultsViewer.setCurrentVarValue(*args, **kwargs)

    def getCurrentVarValue(self):
        return self._resultsViewer.getCurrentVarValue()

    def showResults(self):
        self.setCurrentWidget(self._resultsViewer)

    def _configurationChanged(self, *args, **kwargs):
        self.sigConfigurationChanged.emit()

    def lockAutoFocus(self, lock):
        self._resultsViewer.lockAutoFocus(lock=lock)

    def isAutoFocusLock(self):
        return self._resultsViewer.isAutoFocusLock()

    def hideAutoFocusButton(self):
        self._resultsViewer.hideAutoFocusButton()

    def getDeltaBetaValues(self):
        """Return db values to be computed"""
        return self._deltaBetaSelectionWidget.getDeltaBetaValues()

    def setDeltaBetaValues(self, values):
        """set db values to be computed"""
        self._deltaBetaSelectionWidget.setDeltaBetaValues(values)

    def loadPreprocessingParams(self):
        """load reconstruction nabu if tomwer has already process this
        dataset. Not done for now"""
        return False

    def setScan(self, scan):
        self._resultsViewer.setScan(scan)
        self._nabuSettings.setScan(scan)
        if self.loadPreprocessingParams() and scan.axis_params is not None:
            self._nabuSettings.setConfiguration(scan.axis_params)
        self._deltaBetaSelectionWidget.setScan(scan)

    def getConfiguration(self):
        nabu_config = self.getNabuReconsParams()
        enable_ht = int(self._nabuSettings.getMode() is _NabuMode.HALF_ACQ)
        nabu_config["reconstruction"]["enable_halftomo"] = enable_ht
        # update phase option
        nabu_config["phase"][
            "padding_type"
        ] = self._deltaBetaSelectionWidget.getPaddingType().value
        nabu_config["phase"][
            "unsharp_coeff"
        ] = self._deltaBetaSelectionWidget.getUnsharpCoeff()
        nabu_config["phase"][
            "unsharp_sigma"
        ] = self._deltaBetaSelectionWidget.getUnsharpSigma()
        return {
            "slice_index": self.getReconstructionSlice(),
            "nabu_params": nabu_config,
            "score_method": self.getScoreMethod().value,
            "delta_beta_values": self.getDeltaBetaValues(),
            "output_dir": nabu_config.get("output", {}).get("location", None) or None,
        }

    def getReconstructionSlice(self):
        return self._deltaBetaSelectionWidget.getSlice()

    def setReconstructionSlice(self, slice_):
        return self._deltaBetaSelectionWidget.setSlice(slice_=slice_)

    def setConfiguration(self, config):
        if isinstance(config, QSADeltaBetaParams):
            config = config.to_dict()
        if not isinstance(config, dict):
            raise TypeError(
                f"config should be a dictionary or a SAAxisParams. Not {type(config)}"
            )

        db_values = config.get("delta_beta_values", None)
        if db_values is not None:
            self.setDeltaBetaValues(db_values)
        slice_index = config.get("slice_index", None)
        if slice_index is not None:
            self.setReconstructionSlice(slice_index)
        if "nabu_params" in config:
            nabu_params = config["nabu_params"]
            self.setNabuReconsParams(nabu_params)
            # special handling of unsharp coeff, sigma and padding type
            padding_type = nabu_params.get("phase", {}).get("padding_type", None)
            if padding_type is not None:
                self._deltaBetaSelectionWidget.setPaddingType(padding_type=padding_type)

            unsharp_coeff = nabu_params.get("phase", {}).get("unsharp_coeff", None)
            if unsharp_coeff is not None:
                self._deltaBetaSelectionWidget.setUnsharpCoeff(coeff=unsharp_coeff)

            unsharp_sigma = nabu_params.get("phase", {}).get("unsharp_sigma", None)
            if unsharp_sigma is not None:
                self._deltaBetaSelectionWidget.setUnsharpSigma(sigma=unsharp_sigma)

        if "score_method" in config:
            self.setScoreMethod(config["score_method"])

    def getScoreMethod(self):
        return self._resultsViewer.getScoreMethod()

    def setScoreMethod(self, method):
        self._resultsViewer.setScoreMethod(method)

    def close(self):
        self._resultsViewer.close()
        self._resultsViewer = None
        self._deltaBetaSelectionWidget.close()
        self._deltaBetaSelectionWidget = None
        self._nabuSettings.close()
        self._nabuSettings = None
        super().close()

    def setSliceRange(self, min_, max_):
        self._deltaBetaSelectionWidget.setSliceRange(min_, max_)


class SADeltaBetaWindow(qt.QMainWindow):
    """
    Widget used to determine half-automatically the better delta / beta value
    """

    def __init__(self, parent=None, backend=None):
        qt.QMainWindow.__init__(self, parent)
        self._db_values = []
        self._urls = []

        self._scan = None
        self._qdeltabeta_rp = QSADeltaBetaParams()
        self.setWindowFlags(qt.Qt.Widget)
        # central widget
        self._mainWidget = qt.QWidget(self)
        self._mainWidget.setLayout(qt.QVBoxLayout())
        self._scanInfo = ScanNameLabelAndShape(self)
        self._mainWidget.layout().addWidget(self._scanInfo)
        self._tabWidget = _SADeltaBetaTabWidget(self, backend=backend)
        self._mainWidget.layout().addWidget(self._tabWidget)
        self.setCentralWidget(self._mainWidget)
        # next and previous buttons for browsing the tab widget
        self._browserButtons = TabBrowsersButtons(self)
        self._dockWidgetBrwButtons = qt.QDockWidget(self)
        self._dockWidgetBrwButtons.setWidget(self._browserButtons)
        self.addDockWidget(qt.Qt.BottomDockWidgetArea, self._dockWidgetBrwButtons)
        self._dockWidgetBrwButtons.setFeatures(qt.QDockWidget.DockWidgetMovable)
        # control widget (validate, compute, cor positions)
        self._sadbControl = ControlWidget(self)
        self._dockWidgetCtrl = qt.QDockWidget(self)
        self._dockWidgetCtrl.setWidget(self._sadbControl)
        self.addDockWidget(qt.Qt.BottomDockWidgetArea, self._dockWidgetCtrl)
        self._dockWidgetCtrl.setFeatures(qt.QDockWidget.DockWidgetMovable)

        # expose signals
        self.sigConfigurationChanged = self._tabWidget.sigConfigurationChanged

        # connect signal / slot
        self._browserButtons.sigNextReleased.connect(self._showNextPage)
        self._browserButtons.sigPreviousReleased.connect(self._showPreviousPage)
        self._sadbControl.sigComputationRequest.connect(self._launchReconstructions)
        self._sadbControl.sigValidateRequest.connect(self._validate)

    def showResults(self):
        self._tabWidget.showResults()

    def compute(self):
        """force compute of the current scan"""
        """force compute of the current scan"""
        self._sadbControl.sigComputationRequest.emit()

    def getConfiguration(self) -> dict:
        return self._tabWidget.getConfiguration()

    def setConfiguration(self, config: dict):
        self._tabWidget.setConfiguration(config)

    def getQDeltaBetaRP(self):
        return self._qdeltabeta_rp

    def setScan(self, scan):
        self._scan = scan
        self._tabWidget.setScan(scan)
        self._scanInfo.setScan(scan)

    def getScan(self):
        return self._scan

    def lockAutofocus(self, lock):
        self._tabWidget.lockAutoFocus(lock=lock)

    def isAutoFocusLock(self):
        return self._tabWidget.isAutoFocusLock()

    def hideAutoFocusButton(self):
        return self._tabWidget.hideAutoFocusButton()

    def _launchReconstructions(self):
        """callback when we want to launch the reconstruction of the
        slice for n delta/beta value"""
        raise NotImplementedError("Base class")

    def _validate(self):
        raise NotImplementedError("Base class")

    def _showNextPage(self, *args, **kwargs):
        idx = self._tabWidget.currentIndex()
        idx += 1
        if idx < self._tabWidget.count():
            self._tabWidget.setCurrentIndex(idx)

    def _showPreviousPage(self, *args, **kwargs):
        idx = self._tabWidget.currentIndex()
        idx -= 1
        if idx >= 0:
            self._tabWidget.setCurrentIndex(idx)

    def close(self):
        self._tabWidget.close()
        self._tabWidget = None
        super().close()

    def setDBScores(
        self,
        scores: dict,
        score_method: str | ScoreMethod,
        img_width=None,
        update_only_scores=False,
    ):
        """

        :param scores: cor value (float) as key and
                            tuple(url: DataUrl, score: float) as value
        """
        self._tabWidget.setDeltaBetaScores(
            scores=scores,
            score_method=score_method,
            img_width=img_width,
            update_only_scores=update_only_scores,
        )

    def getScoreMethod(self):
        return self._tabWidget.getScoreMethod()

    def setCurrentDeltaBetaValue(self, value):
        self._tabWidget.setCurrentVarValue(value)

    def getCurrentDeltaBetaValue(self):
        return self._tabWidget.getCurrentVarValue()

    def setSlicesRange(self, min_, max_):
        self._tabWidget.setSliceRange(min_, max_)

    def saveReconstructedSlicesTo(self, output_folder):
        self._tabWidget.saveReconstructedSlicesTo(output_folder=output_folder)


class DeltaBetaSelectionWidget(qt.QWidget):
    """Widget used to select the range of delta / beta to use"""

    sigConfigurationChanged = qt.Signal()
    """emit when configuration change"""

    _DEFAULT_VERTICAL_SLICE_MODE = ("middle", "other")

    def __init__(self, parent=None):
        qt.QWidget.__init__(self, parent)
        self.setWindowFlags(qt.Qt.Widget)
        self.setLayout(qt.QVBoxLayout())

        # slice selection
        self._sliceGB = qt.QGroupBox("slice", self)
        self.layout().addWidget(self._sliceGB)
        self._sliceGB.setLayout(qt.QGridLayout())
        self._label = qt.QLabel("slice", self)
        self._sliceGB.layout().addWidget(self._label, 0, 0, 1, 1)
        sl_spacer = qt.QWidget(self)
        sl_spacer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum)
        self._sliceGB.layout().addWidget(sl_spacer, 0, 5, 1, 1)

        self._defaultSlicesCB = qt.QComboBox(self)
        for mode in self._DEFAULT_VERTICAL_SLICE_MODE:
            self._defaultSlicesCB.addItem(mode)
        self._sliceGB.layout().addWidget(self._defaultSlicesCB, 0, 1, 1, 1)

        self._sliceSelectionQSB = SliceSelector(self, insert=False, invert_y_axis=True)
        self._sliceSelectionQSB.addSlice(value=0, name="Slice", color="green")
        self._sliceSelectionQSB.setFixedSize(qt.QSize(250, 250))
        self._sliceGB.layout().addWidget(self._sliceSelectionQSB, 1, 0, 1, 5)

        # paganin main window
        self._paganinGB = qt.QGroupBox("paganin", self)
        self._paganinGB.setLayout(qt.QVBoxLayout())
        self._mainWindow = _NabuPhaseConfig(self)
        self._mainWindow.setConfigurationLevel(ConfigurationLevel.ADVANCED)
        self._mainWindow._ctfOpts.hide()
        self._paganinGB.layout().addWidget(self._mainWindow)
        self._mainWindow._methodCB.hide()
        self._mainWindow._methodLabel.hide()
        self.layout().addWidget(self._paganinGB)

        # spacer
        widget_spacer = qt.QWidget(self)
        widget_spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        self.layout().addWidget(widget_spacer)

        # connect signal / slot
        self._sliceSelectionQSB.sigSlicesChanged.connect(self._updated)
        self._mainWindow.sigConfChanged.connect(self._updated)
        self._defaultSlicesCB.currentIndexChanged.connect(self._updated)
        self._defaultSlicesCB.currentIndexChanged.connect(self._updateModeCB)

        self.setSlice("middle")

    def _updated(self, *args, **kwargs):
        self.sigConfigurationChanged.emit()

    def setSlice(self, slice_: str | int):
        if isinstance(slice_, str):
            if slice_ != "middle":
                raise ValueError(f"Slice should be 'middle' or an int. Not {slice_}")
            else:
                self.setMode("middle")
        elif isinstance(slice_, int):
            self.setMode("other")
            self._sliceSelectionQSB.setSliceValue("Slice", slice_)
        elif isinstance(slice_, dict) and "Slice" in slice_:
            self.setMode("other")
            self._sliceSelectionQSB.setSliceValue("Slice", slice_["Slice"])
        else:
            raise TypeError(f"slice should be an int or 'middle'. Not {type(slice_)}")

    def setSliceRange(self, min_, max_):
        self._sliceSelectionQSB.setSlicesRange(min_, max_)

    def getMode(self):
        return self._defaultSlicesCB.currentText()

    def setMode(self, mode):
        if mode not in self._DEFAULT_VERTICAL_SLICE_MODE:
            raise ValueError(
                f"mode should be in {self._DEFAULT_VERTICAL_SLICE_MODE}. Not {mode}."
            )
        idx = self._defaultSlicesCB.findText(mode)
        self._defaultSlicesCB.setCurrentIndex(idx)
        self._updateModeCB()

    def _updateModeCB(self):
        self._sliceSelectionQSB.setVisible(self.getMode() == "other")

    def getSlice(self):
        if self.getSliceMode() == "middle":
            return "middle"
        else:
            return self._sliceSelectionQSB.getSlicesValue()

    def getSliceMode(self):
        return self._defaultSlicesCB.currentText()

    def setScan(self, scan: TomwerScanBase):
        self._sliceSelectionQSB.setSlicesRange(0, scan.dim_2)

    def getDeltaBetaValues(self) -> numpy.array:
        db_values = self._mainWindow._paganinOpts.getDeltaBeta()
        return retrieve_lst_of_value_from_str(db_values, type_=float)

    def setDeltaBetaValues(self, values: numpy.array | Iterable):
        values = numpy.array(values)
        step = None
        if len(values) > 3:
            deltas = values[1:] - values[:-1]
            if deltas.min() == deltas.max():
                step = deltas.min()

        deltaBetaQLE = self._mainWindow._paganinOpts._deltaBetaQLE
        if step is None:
            deltaBetaQLE.setText(",".join([str(value) for value in values]))
        else:
            deltaBetaQLE.setText(
                "{from_}:{to_}:{step_}".format(
                    from_=values.min(), to_=values.max(), step_=step
                )
            )

    def getPaddingType(self):
        return self._mainWindow.getPaddingType()

    def setPaddingType(self, padding_type: str):
        self._mainWindow.setPaddingType(padding_type=padding_type)

    def getUnsharpCoeff(self) -> float:
        return self._mainWindow.getUnsharpCoeff()

    def setUnsharpCoeff(self, coeff) -> float:
        return self._mainWindow.setUnsharpCoeff(coeff=coeff)

    def getUnsharpSigma(self) -> float:
        return self._mainWindow.getUnsharpSigma()

    def setUnsharpSigma(self, sigma):
        self._mainWindow.setUnsharpSigma(sigma=sigma)



import os
import pickle

import numpy
import pytest
import pint
from nxtomomill.nexus.nxtomo import NXtomo
from orangecanvas.scheme.readwrite import literal_dumps
from silx.gui.utils.testutils import SignalListener
from nxtomo.nxobject.nxdetector import ImageKey
from orangecontrib.tomwer.widgets.edit.NXtomoEditorOW import NXtomoEditorOW
from tomwer.core.process.edit.nxtomoeditor import NXtomoEditorKeys
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.tests.utils import skip_gui_test
from tomwer.gui.utils.qt_utils import QSignalSpy
from tomwer.tests.conftest import qtapp  # noqa F401
from tomwer.tests.test_utils import is_lr_flip, is_ud_flip

_ureg = pint.get_application_registry()


def getDefaultConfig() -> dict:
    """return the configuration of the NXtomo editor. First value is the value of the field, second is: is the associated lock button locked or not"""
    return {
        NXtomoEditorKeys.ENERGY: (5.9, False),
        NXtomoEditorKeys.SAMPLE_SOURCE_DISTANCE: (-55.2, False),
        NXtomoEditorKeys.PROPAGATION_DISTANCE: (10.1, True),
        NXtomoEditorKeys.SAMPLE_DETECTOR_DISTANCE: (2.4, True),
        NXtomoEditorKeys.FIELD_OF_VIEW: ("Full", False),
        NXtomoEditorKeys.DETECTOR_X_PIXEL_SIZE: (0.023, True),
        NXtomoEditorKeys.DETECTOR_Y_PIXEL_SIZE: (0.025, True),
        NXtomoEditorKeys.SAMPLE_X_PIXEL_SIZE: (0.24, True),
        NXtomoEditorKeys.SAMPLE_Y_PIXEL_SIZE: (0.26, True),
        NXtomoEditorKeys.LR_FLIPPED: (True, True),
        NXtomoEditorKeys.UD_FLIPPED: (False, False),
        NXtomoEditorKeys.X_TRANSLATION: (0.0,),
        NXtomoEditorKeys.Z_TRANSLATION: (0.0,),
    }


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
def test_NXtomoEditorOW(
    qtapp,  # noqa F811
    tmp_path,
):
    window = NXtomoEditorOW()
    # test serialization
    window.setConfiguration(getDefaultConfig())
    assert window.getConfiguration() == getDefaultConfig()
    pickle.dumps(window.getConfiguration())

    # test literal dumps
    literal_dumps(window.getConfiguration())

    # test widget automation
    signal_listener = SignalListener()
    window.sigScanReady.connect(signal_listener)
    # set up the widget to define and lock distance, energy and x pixel size
    distance_widget = window.widget.mainWidget._sampleDetectorDistanceMetricEntry
    distance_widget.setValue(value_m=0.6)
    distance_widget.setUnit(_ureg.millimeter)
    distance_locker = window.widget.mainWidget._sampleDetectorDistanceLB
    distance_locker.setLock(True)
    energy_widget = window.widget.mainWidget._energyEntry
    energy_widget.setValue(88.058)
    energy_locker = window.widget.mainWidget._energyLockerLB
    energy_locker.setLock(True)
    x_pixel_widget = window.widget.mainWidget._xDetectorPixelSizeMetricEntry
    x_pixel_widget.setValue(value_m=45)
    x_pixel_widget.setUnit(_ureg.nanometer)
    x_pixel_locker = window.widget.mainWidget._xDetectorPixelSizeLB
    x_pixel_locker.setLock(True)

    # 1.0 create nx tomos with raw data
    nx_tomo = NXtomo()
    nx_tomo.instrument.detector.x_pixel_size = (
        0.023 * _ureg.meter  # should be overwrite by the configuration / lock buttons
    )
    nx_tomo.instrument.detector.y_pixel_size = (
        0.025 * _ureg.meter  # should be overwrite by the configuration / lock buttons
    )
    nx_tomo.instrument.detector.field_of_view = "full"
    nx_tomo.instrument.detector.distance = (
        2.4 * _ureg.meter  # should be overwrite by the configuration / lock buttons
    )
    nx_tomo.instrument.detector.set_transformation_from_x_flipped(
        False  # should be overwrite by the configuration / lock buttons
    )
    nx_tomo.instrument.detector.set_transformation_from_y_flipped(True)
    nx_tomo.energy = 5.9 * _ureg.keV
    nx_tomo.instrument.detector.image_key_control = [ImageKey.PROJECTION.value] * 12
    nx_tomo.instrument.detector.data = numpy.empty(shape=(12, 10, 10))
    nx_tomo.sample.rotation_angle = numpy.linspace(0, 20, num=12) * _ureg.degree

    file_path = os.path.join(tmp_path, "nxtomo.nx")
    entry = "entry0000"
    nx_tomo.save(
        file_path=file_path,
        data_path=entry,
    )
    # 2.0 set scan to the nxtomo-editor
    scan = NXtomoScan(file_path, entry)
    waiter = QSignalSpy(window.sigScanReady)
    window.setScan(scan=scan)
    # warning: avoid executing the ewoks task as this will be done
    # automatically (has some field lock). This would create concurrency task
    # and could bring some HDF5 concurrency error
    waiter.wait(5000)
    # 3.0 check results are as expected
    # make sure the scan has been re-emitted
    assert signal_listener.callCount() == 1
    # make sure the edition of the parameters have been done and only those
    overwrite_nx_tomo = NXtomo().load(
        file_path=file_path,
        data_path=entry,
        detector_data_as="as_numpy_array",
    )
    numpy.testing.assert_almost_equal(
        overwrite_nx_tomo.instrument.detector.x_pixel_size.to_base_units().magnitude,
        (45e-9 * _ureg.meter).to_base_units().magnitude,
    )
    assert (
        overwrite_nx_tomo.instrument.detector.y_pixel_size
        == nx_tomo.instrument.detector.y_pixel_size
    )
    assert (
        overwrite_nx_tomo.instrument.detector.field_of_view
        == nx_tomo.instrument.detector.field_of_view
    )
    numpy.testing.assert_almost_equal(
        overwrite_nx_tomo.instrument.detector.distance.to_base_units().magnitude,
        (6.0e-4 * _ureg.meter).to_base_units().magnitude,
    )

    assert (
        is_lr_flip(
            overwrite_nx_tomo.instrument.detector.transformations.transformations
        )
        is True
    )
    assert (
        is_ud_flip(
            overwrite_nx_tomo.instrument.detector.transformations.transformations
        )
        is True
    )

    numpy.testing.assert_almost_equal(
        overwrite_nx_tomo.energy.to_base_units().magnitude,
        (88.058 * _ureg.keV).to_base_units().magnitude,
    )
    numpy.testing.assert_array_almost_equal(
        overwrite_nx_tomo.instrument.detector.data,
        nx_tomo.instrument.detector.data,
    )



"""
Define the tomwer Task class.
Insure connection with ewoks.
All instances of tomwer Tasks should avoid gui import (as qt).
"""

from __future__ import annotations

import logging
from collections import namedtuple

from ewokscore.task import Task as _EwoksTask
from ewokscore.taskwithprogress import TaskWithProgress as _EwoksTaskWithProgress


_logger = logging.getLogger(__name__)

_process_desc = namedtuple(
    "_process_desc", ["process_order", "configuration", "results"]
)


class BaseProcessInfo:
    """Tomwer base process class"""

    def __init__(self, inputs=None):
        """
        :param return_dict: if True serialize (to_dict / from_dict functions) between each task
        """

        self._scheme_title = (
            "scheme_title"  # TODO: have a look, this must be get somewhere and reused ?
        )

        """should the return type of the handler should be TomoBase instance
        objects or dict"""
        self._settings = {}
        self._cancelled = False
        # a useful variable that can be set to True if the task has been cancelled

    @staticmethod
    def properties_help():
        """

        :return: display the list of all managed keys and possible values
        """
        # TODO: use argsparse instead of this dict ?
        raise NotImplementedError("BaseProcess is an abstract class")

    @staticmethod
    def program_name():
        """Name of the program used for this processing"""
        raise NotImplementedError("Base class")

    @staticmethod
    def program_version():
        """version of the program used for this processing"""
        raise NotImplementedError("Base class")

    @staticmethod
    def definition():
        """definition of the process"""
        raise NotImplementedError("Base class")

    def get_configuration(self) -> dict | None:
        """

        :return: configuration of the process
        """
        if self._settings is None:
            return None
        if len(self._settings) > 0:
            return self._settings
        else:
            return None

    def set_configuration(self, configuration: dict) -> None:
        self._settings = configuration


class TaskWithProgress(_EwoksTaskWithProgress, BaseProcessInfo):
    """Class from which all tomwer process should inherit

    :param logger: the logger used by the class
    """

    def __init__(
        self,
        varinfo=None,
        inputs=None,
        node_id=None,
        node_attrs=None,
        execinfo=None,
        progress=None,
    ):
        BaseProcessInfo.__init__(self, inputs=inputs)
        _EwoksTaskWithProgress.__init__(
            self,
            varinfo=varinfo,
            inputs=inputs,
            node_id=node_id,
            node_attrs=node_attrs,
            execinfo=execinfo,
            progress=progress,
        )


class Task(_EwoksTask, BaseProcessInfo):
    """Class from which all tomwer process should inherit

    :param logger: the logger used by the class
    """

    def __init__(
        self,
        varinfo=None,
        inputs=None,
        node_id=None,
        node_attrs=None,
        execinfo=None,
    ):
        BaseProcessInfo.__init__(self, inputs=inputs)
        _EwoksTask.__init__(
            self,
            varinfo=varinfo,
            inputs=inputs,
            node_id=node_id,
            node_attrs=node_attrs,
            execinfo=execinfo,
        )



def concatenate_dict(dict_1, dict_2) -> dict:
    """update dict which has dict as values. And we want concatenate those values to"""
    res = dict_1.copy()
    for key in dict_2:
        if key in dict_1:
            if key in [f"axis_{axis}_params" for axis in (0, 1, 2)]:
                res[key] = ";".join((dict_1[key], dict_2[key]))
            elif isinstance(dict_1[key], dict):
                res[key] = concatenate_dict(dict_1=dict_1[key], dict_2=dict_2[key])
            else:
                res[key].update(dict_2[key])
        else:
            res[key] = dict_2[key]
    return res



from __future__ import annotations

import pint
import numpy

from tomwer.utils import Launcher
from nxtomo.nxobject.nxtransformations import (
    get_lr_flip,
    get_ud_flip,
)

_ureg = pint.get_application_registry()


def test_launcher():
    launcher = Launcher(prog="tomwer", version="1.0")
    launcher.add_command(
        "canvas", module_name="tomwer.app.canvas", description="open the orange-canvas"
    )
    launcher.print_help()
    launcher.execute_help(["tomwer", "canvas"])


def is_lr_flip(transformations) -> bool:
    """sum all rotation the detector 'y' axis and check that the sum is 180 (in this case we consider this is a detector left-right flip)"""
    lr_rot_sum = sum(
        [trans.transformation_values for trans in get_lr_flip(transformations)]
    )
    assert isinstance(lr_rot_sum, pint.Quantity)
    return bool(
        numpy.isclose(
            lr_rot_sum,
            180.0 * _ureg.degree,
        )
    )


def is_ud_flip(transformations) -> bool:
    """sum all rotation the detector 'x' axis and check that the sum is 180 (in this case we consider this is a detector up-down flip)"""
    ud_rot_sum = sum(
        [trans.transformation_values for trans in get_ud_flip(transformations)]
    )
    assert isinstance(ud_rot_sum, pint.Quantity)
    return bool(
        numpy.isclose(
            ud_rot_sum,
            180.0 * _ureg.degree,
        )
    )



class HelicalMetadata:
    def __init__(self) -> None:
        self._processes_files = (
            "{scan_parent_dir_basename}/{scan_dir_name}/map_and_doubleff.hdf5"
        )

    @property
    def processes_files(self) -> str:
        return self._processes_files

    @processes_files.setter
    def processes_files(self, path: str):
        if not isinstance(path, str):
            raise TypeError(f"expects a str. Get {path}")
        self._processes_files = path



"""
Deprecated tomwer widgets
"""

NAME = "Deprecated tools"

ID = "orangecontrib.tomwer.widgets.deprecated"

DESCRIPTION = "Some deprecated widgets that will disappear soon"

LONG_DESCRIPTION = DESCRIPTION

ICON = "../../widgets/icons/deprecated.png"

BACKGROUND = "#C0CCFF"

PRIORITY = 99



from __future__ import annotations

import logging
from copy import copy


from nxtomomill.models.edf2nx import EDF2nxModel
from orangewidget import gui
from silx.gui import qt
from silx.gui.utils import blockSignals

from orangecontrib.tomwer.orange.managedprocess import TomwerWithStackStack
from orangecontrib.tomwer.widgets.control.NXTomomillMixIn import NXTomomillMixIn
from tomwer.core.process.control.nxtomomill import EDFToNxProcess
from tomwer.core.process.output import ProcessDataOutputDirMode
from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.gui.control.datalist import EDFDataListMainWindow

logger = logging.getLogger(__name__)


class EDF2NXOW(
    TomwerWithStackStack,
    NXTomomillMixIn,
    ewokstaskclass=EDFToNxProcess,
):
    """
    Widget to allow user to pick some bliss files and that will convert them
    to HDF5scan.
    """

    name = "nxtomomill - edf2nx (Spec-EDF)"
    id = "orange.widgets.tomwer.control.NXTomomillOW.EDF2NXOW"
    description = "Convert folders with .edf to .nx"
    icon = "icons/edf2nx.svg"
    priority = 121
    keywords = [
        "edf",
        "nexus",
        "tomwer",
        "file",
        "convert",
        "NXTomo",
        "tomography",
        "edf2nx",
        "nxtomomill",
    ]

    want_main_area = True
    want_control_area = False
    resizing_enabled = True

    CONFIG_CLS = EDF2nxModel

    LOGGER = logger

    _ewoks_inputs_to_hide_from_orange = (
        "edf_to_nx_configuration",
        "progress",
        "serialize_output_data",
    )

    def __init__(self, parent=None):
        self.__configuration_cache = None
        # cache updated for each folder in order to match `_execute_ewoks_task` design

        TomwerWithStackStack.__init__(self, parent=parent)
        NXTomomillMixIn.__init__(self)
        _layout = gui.vBox(self.mainArea, self.name).layout()
        self.widget = EDFDataListMainWindow(parent=self)
        _layout.addWidget(self.widget)
        # for edf default output path is still the 'near to input file'
        self.widget._dialog._nxTomomillOutputWidget._inScanFolder.setChecked(True)
        # add 'convert auto' check box
        self._convertAutoCB = qt.QCheckBox(
            "convert automatically when edf scan send through 'edf scan' channel", self
        )
        self._convertAutoCB.setChecked(True)
        _layout.addWidget(self._convertAutoCB)

        # connect signal / slot
        self.widget._sendSelectedButton.clicked.connect(self._sendSelected)
        self.widget.sigNXTomoCFGFileChanged.connect(self._saveNXTomoCfgFile)
        self.widget.sigUpdated.connect(self._updateSettings)
        self._convertAutoCB.toggled.connect(self._updateSettings)

        # set default configuration is no existing configuration file defined in the settings
        self.update_default_inputs(
            edf_to_nx_configuration=EDF2nxModel().to_nested_model().model_dump(),
        )

        if isinstance(self.task_output_changed_callbacks, set):
            self.task_output_changed_callbacks.add(self._notify_state)
        elif isinstance(self.task_output_changed_callbacks, list):
            self.task_output_changed_callbacks.append(self._notify_state)
        else:
            raise NotImplementedError

        # handle settings
        self._loadSettings()

    def _updateSettings(self):
        # keep auto conversion setting
        self._ewoks_default_inputs[  # pylint: disable=E1137
            "convert_auto_edfscan_channel"
        ] = self.convertAutoEDFScanReceivedFromChannel()
        # store scans
        self._scans = []
        for scan in self.widget.datalist._myitems:
            self._scans.append(scan)
        # store output folder
        output_dir = self.widget.getOutputFolder()
        if isinstance(output_dir, ProcessDataOutputDirMode):
            output_dir = output_dir.value
        self._ewoks_default_inputs["output_dir"] = output_dir  # pylint: disable=E1137

    def convertAutoEDFScanReceivedFromChannel(self) -> bool:
        return self._convertAutoCB.isChecked()

    def setConvertAutoEDFScanReceivedFromChannel(self, checked: bool):
        self._convertAutoCB.setChecked(checked)

    def _loadSettings(self):
        with blockSignals(self.widget):
            for scan in self._scans:  # pylint: disable=E1133
                self.widget.add(scan)
            if (
                "nxtomomill_cfg_file" in self._ewoks_default_inputs
            ):  # pylint: disable=E1135
                self._nxtomo_cfg_file = (
                    self._ewoks_default_inputs[  # pylint: disable=E1136
                        "nxtomomill_cfg_file"
                    ]
                )
                self.widget.setCFGFilePath(self._nxtomo_cfg_file)
            if "output_dir" in self._ewoks_default_inputs:  # pylint: disable=E1135
                self.widget.setOutputFolder(
                    self._ewoks_default_inputs["output_dir"]  # pylint: disable=E1136
                )
            if (
                "convert_auto_edfscan_channel"
                in self._ewoks_default_inputs  # pylint: disable=E1135
            ):
                self.setConvertAutoEDFScanReceivedFromChannel(
                    self._ewoks_default_inputs[  # pylint: disable=E1136
                        "convert_auto_edfscan_channel"
                    ]
                )

    def _convertAndSend(self, edf_scan: EDFTomoScan):
        # cache updated for each folder in order to match `_execute_ewoks_task` design
        # for edf2nx we need to update input dir from the DataList and output_dir from
        # the one requested by the user
        if not isinstance(edf_scan, EDFTomoScan):
            raise TypeError(
                f"edf_scan is expected to be a {EDFTomoScan} not {type(edf_scan)}"
            )

        self.__configuration_cache = EDF2nxModel.from_dict(
            copy(self.get_default_input_values()["edf_to_nx_configuration"])
        )

        self.__configuration_cache.input_folder = edf_scan.path
        self.__configuration_cache.dataset_basename = edf_scan.dataset_basename

        output_file = EDFToNxProcess.deduce_output_file_path(
            folder_path=edf_scan.path,
            output_dir=self.widget.getOutputFolder(),
            scan=edf_scan,
        )
        output_entry = "entry"
        self.__configuration_cache.output_file = output_file
        # keep 'edf_to_nx_configuration' up to date according to input folder and output_file updates
        self.update_default_inputs(
            edf_to_nx_configuration=self.__configuration_cache.to_dict()
        )

        self.notify_on_going(scan=NXtomoScan(output_file, output_entry))
        try:
            self._execute_ewoks_task(  # pylint: disable=E1123
                propagate=True,
                log_missing_inputs=False,
            )
        except Exception:
            self._execute_ewoks_task(propagate=True)  # pylint: disable=E1123, E1120

    def get_task_inputs(self):
        return {
            "edf_to_nx_configuration": self.__configuration_cache.to_dict(),
            "serialize_output_data": False,
        }

    def handleNewSignals(self) -> None:
        """Invoked by the workflow signal propagation manager after all
        signals handlers have been called.
        """
        # for now we want to avoid propagation any processing.
        # task will be executed only when the user validates the dialog
        edf_scan = super().get_task_inputs().get("edf_scan", None)
        if edf_scan is not None:
            if not isinstance(edf_scan, EDFTomoScan):
                raise TypeError("edf_scan is expected to be an instance of EDFTomoScan")
            self.add(edf_scan.path)
            if self.convertAutoEDFScanReceivedFromChannel():
                self._convertAndSend(edf_scan)

    def _notify_state(self):
        try:
            task_executor = self.sender()
            task_suceeded = task_executor.succeeded
            config = task_executor.current_task.inputs.edf_to_nx_configuration
            config = EDF2nxModel.from_dict(config)
            scan = NXtomoScan(
                config.output_file,
                "entry",
            )
            if task_suceeded:
                self.notify_succeed(scan=scan)
            else:
                self.notify_failed(scan=scan)
        except Exception as e:
            logger.error(f"failed to handle task finished callback. Reason is {e}")

    def _saveNXTomoCfgFile(self, cfg_file):
        super()._saveNXTomoCfgFile(cfg_file, keyword="edf_to_nx_configuration")

    def _sendSelected(self):
        """send all selected items to be converted"""
        self._canOverwriteOutputs = False
        for scan_id in self.widget.datalist._myitems:
            tomo_obj = self.widget.datalist.getEDFTomoScan(scan_id, None)
            if tomo_obj:
                self._convertAndSend(tomo_obj)

    def _sendAll(self):
        """send all items to be converted"""
        self._canOverwriteOutputs = False
        for scan_id in self.widget.datalist._myitems:
            tomo_obj = self.widget.datalist.getEDFTomoScan(scan_id, None)
            if tomo_obj:
                self._convertAndSend(tomo_obj)

    def keyPressEvent(self, event):
        # forward Ctrl+A to the list as the shift ease selection of all
        modifiers = event.modifiers()
        key = event.key()
        if key == qt.Qt.Key_A and modifiers == qt.Qt.KeyboardModifier.ControlModifier:
            self.widget._widget.datalist.keyPressEvent(event)
        super().keyPressEvent(event)



from __future__ import annotations

from silx.gui import qt
from tomwer.gui.configuration.action import (
    ExpertConfigurationAction,
    BasicConfigurationAction,
)
from tomwer.gui.configuration.level import ConfigurationLevel
from tomwer.core.scan.scanbase import TomwerScanBase

try:
    from ewoksnotify.gui.icat import PublishProcessedDataWidget
except ImportError:
    has_ewoksnotify = False
else:
    has_ewoksnotify = True


class PublishProcessedDataWindow(qt.QMainWindow):
    def __init__(self, parent: qt.QWidget | None, beamlines: tuple) -> None:
        super().__init__(parent)
        if not has_ewoksnotify:
            raise RuntimeError(
                f"ewoksnotify not installed. Please  install it to instantiate {PublishProcessedDataWindow}"
            )

        # define toolbar
        toolbar = qt.QToolBar(self)
        self.addToolBar(qt.Qt.TopToolBarArea, toolbar)

        self.__configurationModesAction = qt.QAction(self)
        self.__configurationModesAction.setCheckable(False)
        menu = qt.QMenu(self)
        self.__configurationModesAction.setMenu(menu)
        toolbar.addAction(self.__configurationModesAction)

        self.__configurationModesGroup = qt.QActionGroup(self)
        self.__configurationModesGroup.setExclusive(True)

        self._basicConfigAction = BasicConfigurationAction(toolbar)
        menu.addAction(self._basicConfigAction)
        self.__configurationModesGroup.addAction(self._basicConfigAction)
        self._expertConfiguration = ExpertConfigurationAction(toolbar)
        menu.addAction(self._expertConfiguration)
        self.__configurationModesGroup.addAction(self._expertConfiguration)

        # define widget
        self._centralWidget = PublishProcessedDataWidget(
            parent=parent,
            beamlines=beamlines,
        )
        # hide the dataset field has defined automatically by tomwer
        self._centralWidget._datasetPLB.setDisabled(True)
        self._centralWidget._datasetQLE.setDisabled(True)
        self._centralWidget._datasetQLE.setToolTip("This field is imposed by tomwer")
        self.setCentralWidget(self._centralWidget)

        # connect signal / slot
        self.__configurationModesGroup.triggered.connect(self._userModeChanged)

        # set up
        # for the time beeing let's hide the toolbar and use the 'expert configuration'
        toolbar.hide()
        self._expertConfiguration.trigger()

    def _userModeChanged(self, action):
        self.__configurationModesAction.setIcon(action.icon())
        self.__configurationModesAction.setToolTip(action.tooltip())
        if action is self._basicConfigAction:
            level = ConfigurationLevel.OPTIONAL
        elif action is self._expertConfiguration:
            level = ConfigurationLevel.ADVANCED
        else:
            raise NotImplementedError
        self.centralWidget().setVisible(ConfigurationLevel.ADVANCED <= level)

    # expose API
    def getConfiguration(self) -> dict:
        return self._centralWidget.getConfiguration()

    def setConfiguration(self, configuration: dict) -> None:
        self._centralWidget.setConfiguration(configuration=configuration)

    def setScan(self, scan: TomwerScanBase):
        if not isinstance(scan, TomwerScanBase):
            raise TypeError(
                f"scan is expected to be an instance of {TomwerScanBase}, {type(scan)} provided instead"
            )

        new_config = {}
        if not self._centralWidget._proposalPLB.isChecked():
            new_config["proposal"] = scan.get_proposal_name()
        if not self._centralWidget._beamlinePLB.isChecked():
            new_config["beamline"] = scan.instrument_name

        self.setConfiguration(new_config)



# coding: utf-8
from __future__ import annotations

import os
import shutil
import tempfile

import numpy
import pytest
from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt

from tomwer.core.volume import EDFVolume, HDF5Volume
from tomwer.gui.control.volumeselectorwidget import VolumeSelectorWidget
from tomwer.tests.utils import skip_gui_test


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestVolumeSelector(TestCaseQt):
    """
    Simple test for the VolumeSelector
    """

    def setUp(self):
        super().setUp()
        self.folder = tempfile.mkdtemp()

        self.volume_1 = HDF5Volume(
            file_path=os.path.join(self.folder, "my_volume.hdf5"),
            data_path="entry0000",
            data=numpy.linspace(start=1, stop=100, num=300).reshape((3, 10, 10)),
        )
        self.volume_1.save()

        self.volume_2 = HDF5Volume(
            file_path=os.path.join(self.folder, "my_volume.hdf5"),
            data_path="entry0001",
            data=numpy.linspace(start=1, stop=100, num=400).reshape((4, 10, 10)),
        )
        self.volume_2.save()

        self.volume_3 = EDFVolume(
            folder=os.path.join(self.folder, "vol"),
            data=numpy.linspace(start=1, stop=100, num=500).reshape((5, 10, 10)),
        )
        self.volume_3.save()

        self.widget = VolumeSelectorWidget(parent=None)

    def tearDown(self):
        self.widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.widget.close()
        self.widget = None
        shutil.rmtree(self.folder)
        super().tearDown()

    def test(self):
        for volume in (self.volume_1, self.volume_2, self.volume_3):
            self.widget.add(volume)

        # self.widget.show()
        # self.qapp.exec()
        self.assertEqual(self.widget.dataList.n_data(), 3)
        self.widget.remove(self.volume_2)
        self.assertEqual(self.widget.dataList.n_data(), 2)
        self.widget.remove(self.volume_3)
        self.assertEqual(self.widget.dataList.n_data(), 1)
        assert self.widget.dataList.rowCount() == 1



# coding: utf-8

"""module defining dedicated completer"""

from __future__ import annotations


import pytest
from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt

from tomwer.gui.utils.completer import UrlCompleterDialog
from tomwer.tests.utils import skip_gui_test


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestUrlCompleterDialog(TestCaseQt):
    def setUp(self):
        super().setUp()
        self._urls = (
            "test",
            "test1",
            "test2",
        )
        self._dialog = UrlCompleterDialog(
            urls=self._urls,
            current_url="test",
        )

    def tearDown(self):
        self._dialog.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._dialog.close()
        self._dialog = None
        super().tearDown()

    def test(self):
        """simple test on the dialog behavior"""
        assert self._dialog._buttons.button(qt.QDialogButtonBox.Ok).isEnabled()
        self._dialog._completerWidget.setText("toto")
        self.qapp.processEvents()
        assert not self._dialog._buttons.button(qt.QDialogButtonBox.Ok).isEnabled()



"""Some widget construction to check if a sample moved"""

from __future__ import annotations

import logging

import sys
import h5py
from silx.gui import qt
from silx.io.url import DataUrl

from .metadataloader import VolumeMetadataLoader
from tomwer.core.tomwer_object import TomwerObject
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.volume.volumefactory import VolumeFactory
from tomwer.core.volume.volumebase import TomwerVolumeBase
from tomwer.io.utils.tomoobj import get_tomo_objs_instances
from tomwer.gui.dialog.QDataDialog import QDataDialog
from tomwer.gui.stackplot import QImageFileStackPlot
from tomwer.gui.utils.loadingmode import LoadingModeToolButton, LoadingMode

_logger = logging.getLogger(__name__)


class _ImageStack(qt.QMainWindow):
    """
    Base class to :class:`SliceStack` and :class:`RadioStack` classes.
    Define the common layout and interaction.
    """

    METADATA_LOADER_CLASS = None

    SHOW_SCAN_OVERVIEW = True

    SHOW_VOLUME_METADATA_OVERVIEW = True

    sigMetadataLoaded = qt.Signal()

    def __init__(self, parent=None, backend=None):

        super().__init__(parent)

        self.setWindowFlags(qt.Qt.Widget)
        self._viewer = QImageFileStackPlot(
            parent=self, backend=backend, show_overview=self.SHOW_SCAN_OVERVIEW
        )
        if not self.SHOW_VOLUME_METADATA_OVERVIEW:
            self._viewer._reconsInfoDockWidget.hide()
        self.setCentralWidget(self._viewer)

        self._actionsToolbar = qt.QToolBar()
        self.addToolBar(qt.Qt.RightToolBarArea, self._actionsToolbar)

        self._clearAction = _ClearAction(parent=self._actionsToolbar)
        self._actionsToolbar.addAction(self._clearAction)
        self._clearAction.triggered.connect(self.clear)

        self._addTomoObjAction = _AddTomoObjectAction(parent=self._actionsToolbar)
        self._actionsToolbar.addAction(self._addTomoObjAction)
        self._addTomoObjAction.triggered.connect(self._callbackAddNewTomoObjFrmDialog)

        self._loadingMode = LoadingModeToolButton(parent=self)
        self._actionsToolbar.addWidget(self._loadingMode)

        # connect signal / slot
        self.sigMetadataLoaded.connect(self._viewer._updateUrlInfos)
        self._loadingMode.sigLoadModeChanged.connect(self.setLoadingMode)

    def addTomoObj(self, tomo_obj: TomwerObject | str):
        raise NotImplementedError

    def _callbackAddNewTomoObjFrmDialog(self) -> None:

        dialog = QDataDialog(self, multiSelection=True)
        dialog.setNameFilters(
            [
                "Any files (*)",
                "HDF5 files (*.h5 *.hdf5 *.nx *.nxs *.nexus)",
                "Nexus files (*.nx *.nxs *.nexus)",
            ]
        )

        if not dialog.exec():
            dialog.close()
            return

        self._addTomoObjectsFromStrList(dialog.files_selected())

    def _addTomoObjectsFromStrList(self, str_list: list[str]) -> None:
        tomo_objects, _ = get_tomo_objs_instances(str_list)
        for tomo_obj in tomo_objects:
            self.addTomoObj(tomo_obj)

    def clear(self):
        self._viewer.reset()

    def close(self) -> None:
        # clearing free the loading thread. Safer to remove them when closing
        self.clear()
        super().close()

    def setLoadingMode(self, mode: LoadingMode | str) -> None:
        mode = LoadingMode(mode)
        if mode is LoadingMode.LAZY_LOADING:
            # set to the default N_PRELOAD value
            n_prefetch = self._viewer.N_PRELOAD
        else:
            n_prefetch = sys.maxsize
        self._viewer.setNPrefetch(n_prefetch)


class _ClearAction(qt.QAction):
    def __init__(self, parent):
        style = qt.QApplication.instance().style()
        icon = style.standardIcon(qt.QStyle.SP_DialogResetButton)
        qt.QAction.__init__(self, icon, "Clear", parent)


class _AddTomoObjectAction(qt.QAction):
    def __init__(self, parent):
        style = qt.QApplication.instance().style()
        icon = style.standardIcon(qt.QStyle.SP_DirIcon)
        qt.QAction.__init__(self, icon, "Add tomo object", parent)


class SliceStack(_ImageStack):
    """
    Widget displaying all slices contained in a list of acquisition folder
    """

    SHOW_SCAN_OVERVIEW = False

    SHOW_VOLUME_METADATA_OVERVIEW = True

    METADATA_LOADER_CLASS = VolumeMetadataLoader

    def __init__(self, parent=None, backend=None):
        super().__init__(parent, backend)
        self._loadingMetadataThreads = []
        "list of thread processing metadata loading"
        self._volume_id_to_urls: dict[str, tuple[DataUrl]] = {}
        # allow user to edit the list (user can remove an url)
        url_list = self._viewer._urlsTable._urlsTable
        url_list.setEditable(True)
        # allow multiple selection (more convenient when users want to remove a set of url)
        # not for now: seems to have a bug on it
        # url_list.setSelectionMode(qt.QAbstractItemView.MultiSelection)

    @property
    def url_to_volume_id(self, volume_id) -> str | None:
        urls_to_volume_id = {
            (value, key) for key, value in self._volume_id_to_urls.items()
        }
        return urls_to_volume_id.get(volume_id, None)

    def getImagesUrls(self) -> dict[str, dict]:
        """
        Parse all self._scans and find the images to be displayed on the widget

        :return: images to display for each scan. Key is the volume identifier, value is a dict
                 with index as key and DataUrl as value
        """
        slices = {}
        for volume in self._tomoObjects.values():
            assert isinstance(volume, TomwerVolumeBase)
            slices[volume.get_identifier().to_str()] = tuple(volume.browse_data_urls())
        return slices

    def __tomoObjIdToInstance(self, tomo_obj: TomwerObject | str) -> TomwerObject:
        """Convert a tomo obj to it TomwerObject"""
        if isinstance(tomo_obj, str):
            try:
                tomo_obj = VolumeFactory.create_tomo_object_from_identifier(tomo_obj)
            except (ValueError, TypeError) as e1:
                try:
                    tomo_obj = ScanFactory.create_tomo_object_from_identifier(tomo_obj)
                except (ValueError, TypeError) as e2:
                    raise ValueError(
                        f"Unable to create a tomo object from {tomo_obj}. Errors are {e1}, {e2}"
                    )
        return tomo_obj

    def addTomoObj(self, tomo_obj: TomwerObject | str):
        tomo_obj = self.__tomoObjIdToInstance(tomo_obj=tomo_obj)

        if isinstance(tomo_obj, TomwerScanBase):
            reconstructed_volumes = (
                tomo_obj.latest_reconstructions or tomo_obj.get_reconstructed_slices()
            )
            for volume_id in reconstructed_volumes:
                if isinstance(volume_id, str):
                    volume_id_as_str = volume_id
                elif isinstance(volume_id, TomwerVolumeBase):
                    volume_id_as_str = volume_id.get_identifier().to_str()
                else:
                    volume_id_as_str = volume_id.to_str()
                self.addTomoObj(tomo_obj=volume_id_as_str)
        elif isinstance(tomo_obj, TomwerVolumeBase):
            self._addVolume(volume=tomo_obj)
            # set the latest slice as the active one
            try:
                active_url = next(tomo_obj.browse_data_urls())
            except StopIteration:
                pass
            else:
                try:
                    self._viewer.setCurrentUrl(active_url)
                except Exception:
                    # the url might not exist or might have been removed.
                    # Because each tome we add an object the urls are reset
                    # FIXME: in this case the urls must be updated instead of resetting
                    pass
        else:
            raise TypeError(f"tomo obj type is not handled ({type(tomo_obj)})")

    def _addVolume(self, volume: TomwerVolumeBase):
        new_urls = [url.path() for url in volume.browse_data_urls()]
        self._volume_id_to_urls[volume.get_identifier().to_str()] = tuple(new_urls)
        self._loadMetadata(volume)

        # work around. See issue: TODO add silx issue
        if self._viewer._urlIndexes is None:
            current_urls = tuple()
        else:
            current_urls = self._viewer.getUrls() or tuple()
        updated_urls = {*new_urls} | {*current_urls}
        updated_urls = tuple([DataUrl(path=url_path) for url_path in updated_urls])
        self._viewer.setUrls(updated_urls)

    def setCurrentUrl(self, url: DataUrl):
        if url in ("", None):
            url = None
        elif isinstance(url, str):
            url = DataUrl(path=url)
        elif not isinstance(url, DataUrl):
            raise TypeError
        if url is None:
            pass
            # FIXME: add a function to clear the metadata widget
        super().setCurrentUrl(url)

    # metadata loading
    def _load(self, url: DataUrl):
        # load the frame
        super()._load()
        # load metadata
        url_path = url.path()
        volume_id = self.url_to_volume_id(url_path)
        if volume_id is None:
            _logger.warning(f"Failed to find volume associated to {url_path}")
            return

        volume = VolumeFactory.create_tomo_object_from_identifier(volume_id)
        if volume not in self._loadingMetadataThreads:
            # if a thread is already loading metadata for this volume avoid launching another thread for it
            self._loadMetadata(tomo_obj=volume)

    def _loadMetadata(self, tomo_obj: TomwerObject):
        if not isinstance(tomo_obj, TomwerObject):
            raise TypeError(
                f"volume should be an instance of {TomwerObject}. Got {type(tomo_obj)} instead"
            )
        loader = self.METADATA_LOADER_CLASS(
            tomo_obj_identifier=tomo_obj.get_identifier().to_str(), parent=self
        )
        loader.finished.connect(self._metadataLoaded, qt.Qt.QueuedConnection)
        self._loadingMetadataThreads.append(loader)
        loader.start()

    def _metadataLoaded(self):
        sender = self.sender()
        if sender in self._loadingMetadataThreads:
            self._loadingMetadataThreads.remove(sender)
        self._viewer.updateSliceMetadata(
            {url.path(): sender.metadata for url in sender.volume.browse_data_urls()}
        )
        self.sigMetadataLoaded.emit()

    def _freeMetadataLoadingThreads(self):
        for thread in self._loadingMetadataThreads:
            thread.blockSignals(True)
            thread.wait(50)
        self._loadingMetadataThreads.clear()

    def close(self):
        # clearing free the loading thread. Safer to remove them when closing
        self._freeMetadataLoadingThreads()
        super().close()

    def clear(self):
        self._volume_id_to_urls.clear()
        self._freeMetadataLoadingThreads()
        super().clear()


class RadioStack(_ImageStack):
    """
    Widget displaying all radio contained in a list of acquisition folder
    """

    SHOW_SCAN_OVERVIEW = True

    SHOW_VOLUME_METADATA_OVERVIEW = False

    def getImagesUrls(self) -> dict:
        """
        Parse all self._scans and find the images to be displayed on the widget

        :return: images to display for each scan
        """
        slices = {}
        for scan in self._tomoObjects.values():
            # manage hdf5
            if isinstance(scan, TomwerScanBase):
                imgs = scan.projections
                if len(imgs) > 0:
                    slices[str(scan)] = imgs
            elif h5py.is_hdf5(scan):
                try:
                    scans = ScanFactory.create_scan_objects(scan_path=scan)
                except Exception as e:
                    _logger.warning(e)
                else:
                    for scan_ in scans:
                        imgs = scan_.projections
                        if len(imgs) > 0:
                            slices[str(scan_)] = imgs
            else:
                # manage edf
                try:
                    scan_ = ScanFactory.create_scan_object(scan_path=scan)
                except ValueError:
                    pass
                else:
                    imgs = scan_.projections
                    if len(imgs) > 0:
                        slices[scan] = imgs
        return slices

    def addTomoObj(self, tomo_obj: TomwerScanBase | str):
        if isinstance(tomo_obj, str):
            try:
                tomo_obj = ScanFactory.create_tomo_object_from_identifier(tomo_obj)
            except (ValueError, TypeError) as e:
                _logger.error(
                    f"Fail to create a scan object from {tomo_obj}. Error is {e}"
                )
        if not isinstance(tomo_obj, TomwerScanBase):
            raise TypeError(
                f"tomo_obj should be an instance of {TomwerScanBase}. Get {type(tomo_obj)}"
            )
        self._addScan(scan=tomo_obj)
        self._viewer.updateScanToUrl(
            {url.path(): tomo_obj for url in tomo_obj.projections.values()}
        )

    def _addScan(self, scan: TomwerScanBase):
        new_urls = tuple([url.path() for url in scan.projections.values()])

        # work around. See issue: TODO add silx issue
        if self._viewer._urlIndexes is None:
            current_urls = tuple()
        else:
            current_urls = self._viewer.getUrls() or tuple()
        updated_urls = {*new_urls} | {*current_urls}
        updated_urls = tuple([DataUrl(path=url_path) for url_path in updated_urls])
        self._viewer.setUrls(updated_urls)



from __future__ import annotations

import logging
import os

from silx.gui import icons as silxicons
from silx.gui import qt

from tomwer.core.process.control.datawatcher.datawatcher import _DataWatcher
from tomwer.core.scan.blissscan import BlissScan
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.utils import logconfig
from tomwer.gui.control import datareacheractions, history, observations
from tomwer.gui.control.datawatcher import configuration
from tomwer.gui.control.datawatcher.controlwidget import ControlWidget
from tomwer.gui.utils.waiterthread import QWaiterThread
from tomwer.io.utils import get_default_directory

from .datawatcherobserver import _QDataWatcherObserver, _QOngoingObservation

logger = logging.getLogger(__name__)


class DataWatcherWidget(_DataWatcher, qt.QMainWindow):
    """
    Widget used to display the on-going acquisitions and the finished one.

    :param parent: the parent widget
    """

    obsStatusToWidgetStatus = {
        "not processing": "Not processed",
        "none found": "Running",
        "starting": "Running",
        "started": "Running",
        "waiting for acquisition ending": "Running",
        "acquisition ended": "Executed",
        "acquisition canceled": "Failed",
        "failure": "Failed",
    }

    _animatedStates = (
        "none found",
        "parsing",
        "waiting for acquisition ending",
        "starting",
        "started",
    )

    DEFAULT_DIRECTORY = "/lbsram/data/visitor"

    sigTMStatusChanged = qt.Signal(str)
    """Signal emitted when the state changed"""
    sigScanReady = qt.Signal(object)
    """Signal emitted when a scan is considered as ready"""
    sigFolderObservedChanged = qt.Signal()
    """Signal emitted when the user change the observed folder"""
    sigObservationStart = qt.Signal()
    """Signal emitted when the observation starts"""
    sigObservationEnd = qt.Signal()
    """Signal emitted when the observation end"""
    sigFilterFileNamePatternChanged = qt.Signal(str)
    """Signal emut when the filter pattern change"""

    def __init__(self, parent=None):
        qt.QMainWindow.__init__(self, parent)
        self.setWindowFlags(qt.Qt.Widget)
        self._maxAdv = 100  # maximal progress bar advancement
        self._configWindow = None
        """Widget containing the configuration of the watcher"""
        self._historyWindow = None
        """Widget containing the latest valid scan found by the watcher"""
        self._observationWidget = None
        """Widget containing the current observed directory by the watcher"""
        self.controlWidget = None
        """Widget containing the 'control' of the datawatcher: select folder,
        status ..."""
        _DataWatcher.__init__(self)
        self.setFolderObserved(self.folderObserved)
        self._setStateIcon(silxicons.getQIcon("remove"))

        # connect signal / slot
        self._qteFolderSelected.textChanged.connect(self._updateFolderObserved)
        self._qteFolderSelected.editingFinished.connect(self._restartObservation)
        self._filterQLE.editingFinished.connect(self._restartObservation)
        self._filterQLE.editingFinished.connect(self._filterChanged)
        self._qpbstartstop.released.connect(self._switchObservation)

        # expose api
        self.sigObservationModeChanged = (
            self.getConfigWindow()._observationMethod.sigSelectionChanged
        )

    def _initClass(self):
        self.loopObservationThread = None
        _DataWatcher._initClass(self)
        toolbar = qt.QToolBar("", parent=self)
        toolbar.setIconSize(qt.QSize(32, 32))
        self._controlAction = datareacheractions.ControlAction(parent=self)
        self._observationsAction = datareacheractions.ObservationAction(parent=self)
        self._configurationAction = datareacheractions.ConfigurationAction(parent=self)
        self._historyAction = datareacheractions.HistoryAction(parent=self)
        toolbar.addAction(self._controlAction)
        toolbar.addAction(self._observationsAction)
        toolbar.addAction(self._configurationAction)
        toolbar.addAction(self._historyAction)

        self._actionGroup = qt.QActionGroup(self)
        self._actionGroup.addAction(self._controlAction)
        self._actionGroup.addAction(self._observationsAction)
        self._actionGroup.addAction(self._configurationAction)
        self._actionGroup.addAction(self._historyAction)

        self.addToolBar(qt.Qt.LeftToolBarArea, toolbar)
        toolbar.setMovable(False)

        self._buildGUI()

        # set initial path to observe
        self.setFolderObserved(self._getInitPath())

        # hide all windows by default
        for widget in (
            self.getControlWindow(),
            self.getObservationWidget(),
            self.getConfigWindow(),
            self.getHistoryWindow(),
        ):
            widget.setVisible(False)

        # deal with toolbar connection
        self._controlAction.toggled[bool].connect(self.getControlWindow().setVisible)
        self._observationsAction.toggled[bool].connect(
            self.getObservationWidget().setVisible
        )
        self._configurationAction.toggled[bool].connect(
            self.getConfigWindow().setVisible
        )
        self._historyAction.toggled[bool].connect(self.getHistoryWindow().setVisible)

        self._controlAction.setChecked(True)
        # expose signal
        self.sigObservationModeChanged = (
            self.getConfigWindow()._observationMethod.sigSelectionChanged
        )

    def close(self):
        logger.info("closing the datawatacher")
        self.stop()
        if self.loopObservationThread is not None:
            self.loopObservationThread.blockSignals(True)
            self._disconnectObserverThread()
            self.loopObservationThread.wait(4000)
            self.loopObservationThread = None
        if (
            self.observationThread is not None
            and self.observationThread.isRunning()  # pylint: disable=E1101
        ):
            self.observationThread.blockSignals(True)  # pylint: disable=E1101
            self.observationThread.wait(4000)
            self.observationThread = None
        super(DataWatcherWidget, self).close()

    def getConfigWindow(self):
        if self._configWindow is None:
            self._configWindow = configuration._DWConfigurationWidget(parent=self)
            self._configWindow.startByOldestStateChanged.connect(self.setStartByOldest)
            self._configWindow.startByOldestStateChanged.connect(
                self._restartObservation
            )
        return self._configWindow

    def getObservationWidget(self):
        if self._observationWidget is None:
            self._observationWidget = observations.ScanObservation(parent=self)
            if self.observationThread:
                self._observationWidget.setOnGoingObservations(
                    self.observationThread.observations
                )
        return self._observationWidget

    def _changeCentralWidget(self, widget, action):
        actions = (
            self._historyAction,
            self._observationsAction,
            self._configurationAction,
            self._controlAction,
        )
        for _action in actions:
            _action.blockSignals(True)

        for _action in actions:
            if _action != action:
                _action.setChecked(False)

        _widgets = (
            self.getControlWindow(),
            self.getObservationWidget(),
            self.getHistoryWindow(),
            self.getConfigWindow(),
        )
        for _widget in _widgets:
            _widget.setVisible(False)

        widget.setVisible(action.isChecked())
        for _action in actions:
            _action.blockSignals(False)

    def getControlWindow(self):
        if self.controlWidget is None:
            self.controlWidget = ControlWidget(parent=self)

        return self.controlWidget

    def _buildGUI(self):
        """Build the GUI of the widget"""
        self._mainWidget = qt.QWidget(parent=self)
        self._mainWidget.setLayout(qt.QVBoxLayout())
        self._mainWidget.layout().addWidget(self.getControlWindow())
        self._mainWidget.layout().addWidget(self.getConfigWindow())
        self._mainWidget.layout().addWidget(self.getObservationWidget())
        self._mainWidget.layout().addWidget(self.getHistoryWindow())
        self.setCentralWidget(self._mainWidget)

    def _filterChanged(self, *args, **kwargs):
        self.sigFilterFileNamePatternChanged.emit(self._filterQLE.text())

    def getFilterLinuxFileNamePattern(self) -> str | None:
        value = self._filterQLE.text().rstrip(" ").lstrip(" ")
        if value == "":
            return None
        else:
            return value

    def _buildLoopTimeBreak(self):
        """
        Build the spin box to define the break we want to make between two
        observations
        """
        widget = qt.QWidget(self.controlWidget)
        layout = qt.QHBoxLayout()
        widget.setLayout(layout)

        layout.addWidget(qt.QLabel("Waiting time between observations (in s)"))
        self._qsbLoopTimeBreak = qt.QSpinBox(parent=widget)
        self._qsbLoopTimeBreak.setMinimum(1)
        self._qsbLoopTimeBreak.setMaximum(1000000)
        self._qsbLoopTimeBreak.setValue(self.maxWaitBtwObsLoop)
        self._qsbLoopTimeBreak.valueChanged.connect(self.setWaitTimeBtwLoop)
        layout.addWidget(self._qsbLoopTimeBreak)

        return widget

    @property
    def _qlInfo(self):
        return self.getControlWindow()._qlInfo

    @property
    def _qteFolderSelected(self):
        return self.getControlWindow()._qteFolderSelected

    @property
    def _filterQLE(self):
        return self.getControlWindow()._filterQLE

    @property
    def _qpbstartstop(self):
        return self.getControlWindow()._qpbstartstop

    @property
    def _stateLabel(self):
        return self.getControlWindow()._stateLabel

    @property
    def mystatusBar(self):
        return self.getControlWindow().mystatusBar

    def _getInitPath(self):
        initPath = ""
        if "DATADIR" in os.environ:
            initPath = os.environ["DATADIR"]
            self.getControlWindow()._qlInfo.setText(
                "note : environment variable DATADIR found, "
                "$DATADIR setted has the root of the observe folder"
            )
            myFont = self._qlInfo.font()
            myFont.setItalic(True)
            self._qlInfo.setFont(myFont)
        else:
            self._qlInfo.setText(
                "note : no DATADIR environment variable setted. "
                "Can't set a default root directory for observation"
            )
            myFont = self._qlInfo.font()
            myFont.setItalic(True)
            self._qlInfo.setFont(myFont)
        return initPath

    def getHistoryWindow(self):
        if self._historyWindow is None:
            self._historyWindow = history.ScanHistory(parent=self)
        return self._historyWindow

    def stop(self, sucess=False):
        """
        Stop the thread of observation

        :param sucess: if True this mean that we are stopping the
                            observation because we found an acquisition
                            finished. In this case we don't want to update the
                            status and the log message.

        :return bool: True if the observation have been stopped. Otherwise this
            mean that not observation was executing
        """
        if self.isObserving is False:
            return False

        self._setIsObserving(False)
        if self.loopObservationThread is not None:
            self.loopObservationThread.wait(int(self.maxWaitBtwObsLoop + 2))
        if self.observationThread is not None:
            # remove connection
            self._disconnectObserverThread()
            self.observationThread.blockSignals(True)  # pylint: disable=E1101
            self.observationThread.wait(4000)

        if sucess is False:
            self._setCurrentStatus(str("not processing"))

            if sucess is False:
                if self._observationWidget is not None:
                    self._observationWidget.clear()
                message = "observation stopped"
                self.sigObservationEnd.emit()
                logger.inform(message)
                self.mystatusBar.showMessage(message)
                self._setCurrentStatus("not processing")
            return True

    def start(self):
        """
        Start the thread of observation

         :return bool: True if the observation was started. Otherwise this
            mean that an observation was already running
        """
        if _DataWatcher.start(self):
            mess = f"start observation on {self.folderObserved}"
            logger.inform(mess)
            self.mystatusBar.showMessage(mess)
            self._setCurrentStatus("started")
            self.sigObservationStart.emit()
            return True
        else:
            return False

    def _setIsObserving(self, b):
        _DataWatcher._setIsObserving(self, b)
        if self.isObserving is True:  # pylint: disable=E1101
            self._qpbstartstop.setText(self.getControlWindow()._TXT_STOP_OBS)
        else:
            self._qpbstartstop.setText(self.getControlWindow()._TXT_START_OBS)

    def _setFolderPath(self):  # pragma: no cover
        """
        Ask the user the path to the folder to observe
        """
        defaultDirectory = self.getFolderObserved()
        if defaultDirectory is None or not os.path.isdir(defaultDirectory):
            if os.path.isdir(self.DEFAULT_DIRECTORY):
                defaultDirectory = self.DEFAULT_DIRECTORY
            if defaultDirectory is None:
                defaultDirectory = get_default_directory()

        dialog = qt.QFileDialog(self, directory=defaultDirectory)
        dialog.setFileMode(qt.QFileDialog.DirectoryOnly)

        if not dialog.exec():
            dialog.close()
            return

        self.setFolderObserved(dialog.selectedFiles()[0])

        if self.isObserving:
            self._restartObservation()

    def _updateFolderObserved(self, txt):
        self.folderObserved = self._qteFolderSelected.text()
        self.sigFolderObservedChanged.emit()

    def setFolderObserved(self, path):
        if path is not None and os.path.isdir(path):
            super(DataWatcherWidget, self).setFolderObserved(os.path.abspath(path))
            self._qteFolderSelected.setText(self.folderObserved)

    def setStartByOldest(self, b):
        """
        Set if we want to start parsing files from the oldest or the newest

        :param b:
        """
        _DataWatcher.setStartByOldest(self, b)
        self.getConfigWindow()._qcboldest.setChecked(self.startByOldest)

    def _initObservation(self):
        """
        Init the thread running the data watcher functions
        """
        if _DataWatcher._initObservation(self) is True:
            if self._observationWidget is not None:
                self._observationWidget.setOnGoingObservations(
                    self.observationThread.observations
                )
            return True
        else:
            return False

    def _messageNotDir(self, dir_):
        super()._messageNotDir(dir_=dir_)
        message = f"Given path ({dir_}) isn't a directory."
        self.mystatusBar.showMessage("!!! " + message + "!!!")

    def informationReceived(self, info):
        self.mystatusBar.showMessage(info)

    def _scanStatusChanged(self, scan, status):
        mess = f"scan {os.path.basename(scan)} is observed. Status: {status}"
        self.mystatusBar.showMessage(mess)

    def _connectObserverThread(self):
        if self.observationThread is not None and self.obsThIsConnected is False:
            self.observationThread.observations.sigObsStatusReceived.connect(
                self._scanStatusChanged
            )
            self.observationThread.sigScanReady.connect(self._signalScanReady)
            self.obsThIsConnected = True

    def _disconnectObserverThread(self):
        if self.observationThread is not None and self.obsThIsConnected is True:
            self.observationThread.observations.sigObsStatusReceived.disconnect(
                self._scanStatusChanged
            )
            self.observationThread.sigScanReady.disconnect(self._signalScanReady)
            self.obsThIsConnected = False

    def _updateStatusView(self):
        """Update the processing state"""
        if self.currentStatus in self._animatedStates:
            pass
        elif self.currentStatus == "acquisition ended":
            self._setStateIcon(silxicons.getQIcon("selected"))
        elif self.currentStatus == "failure":
            self._setStateIcon(silxicons.getQIcon("remove"))
        elif self.currentStatus == "not processing":
            self._setStateIcon(None)

    def _setStateIcon(self, icon):
        """set the icon pass in parameter to the state label

        :param icon:the icon to set"""
        # needed for heritage from DataWatcher
        if icon is None:
            self._stateLabel.setPixmap(qt.QIcon().pixmap(30, state=qt.QIcon.On))
        else:
            self._stateLabel.setPixmap(icon.pixmap(30, state=qt.QIcon.On))

    def _signalScanReady(self, scan):
        if type(scan) is str:
            try:
                _scan = ScanFactory.create_scan_object(scan_path=scan)
            except Exception as e:
                logger.error(
                    f"Fail to create a TomoBase instance from {scan} Reason is {e}"
                )
                return
        else:
            assert isinstance(scan, (TomwerScanBase, BlissScan))
            _scan = scan

        _DataWatcher._signalScanReady(self, _scan)
        self.lastFoundScans.add(_scan)
        self._updateLastReceived()

    def _updateLastReceived(self):
        """
        For now we are updating each time the list.
        It would be better to update it instead.
        """
        self.getHistoryWindow().update(scans=list(self.lastFoundScans.items()))

    def _getObservationClass(self):
        return _QOngoingObservation

    def _createDataWatcher(self):
        self.observationThread = _QDataWatcherObserver(
            observationClass=self._getObservationClass(),
            obsMethod=self.obsMethod,
            srcPattern=self.srcPattern,
            destPattern=self.destPattern,
            file_name_pattern=self.linux_filter_file_pattern,
        )
        self._connectObserverThread()

    def _check_finished_scan(self):
        # nothing to do: managed by signals
        pass

    def waitForObservationFinished(self):
        if self.observationThread.observations is not None:
            for _, thread in self.observationThread.observations.dict.items():
                thread.wait(10)

    def _launchObservation(self):
        """Main function of the widget"""
        if self.isObserving is False:
            return

        # manage data watcher observation
        if (
            self.observationThread is None
            or not self.observationThread.isRunning()  # pylint: disable=E1101
        ):
            if self._initObservation() is False:
                self._setCurrentStatus("failure")
                logger.info(
                    "failed on observation",
                    extra={logconfig.DOC_TITLE: self._scheme_title},
                )
                return

        # starting the observation thread
        self.observationThread.start()

        # manage observation loop
        if self.loopObservationThread is None:
            self.loopObservationThread = QWaiterThread(
                self.getTimeBreakBetweenObservation()
            )
            self.loopObservationThread.finished.connect(self._launchObservation)

        if not self.loopObservationThread.isRunning():
            self._connectObserverThread()
            self.loopObservationThread.start()

    def _observation_thread_running(self):
        return (
            self.observationThread is not None
            and self.observationThread.isRunning()  # pylint: disable=E1101
        )



from __future__ import annotations


from typing import Iterable

import numpy
from enum import Enum as _Enum

from tomwer.core.process.reconstruction.scores.params import SABaseParams


class ReconstructionMode(_Enum):
    VERTICAL = "Vertical"
    TILT_CORRECTION = "Tilt correction"


class SAAxisParams(SABaseParams):
    """Parameters for the semi-automatic axis calculation"""

    def __init__(self):
        super().__init__()
        self._research_width = 10  # in pixel
        self._estimated_cor = None
        self._n_reconstruction = 20
        self._mode = ReconstructionMode.VERTICAL
        self._image_width = None
        self._cluster_config = None

    @property
    def research_width(self):
        return self._research_width

    @research_width.setter
    def research_width(self, research_width):
        self._research_width = research_width

    @property
    def estimated_cor(self):
        return self._estimated_cor

    @estimated_cor.setter
    def estimated_cor(self, estimated_cor):
        self._estimated_cor = estimated_cor

    @property
    def cors(self) -> Iterable:
        return self.compute_cors(
            estimated_cor=self.estimated_cor,
            research_width=self.research_width,
            n_reconstruction=self.n_reconstruction,
        )

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, mode):
        mode = ReconstructionMode(mode)
        self._mode = mode

    @property
    def image_width(self) -> float | None:
        return self._image_width

    @image_width.setter
    def image_width(self, width):
        if not isinstance(width, (type(None), float, int)):
            raise TypeError(f"None, int or float expected. Not {type(width)}")
        else:
            self._image_width = width

    def to_dict(self):
        ddict = super().to_dict()
        ddict.update(
            {
                "estimated_cor": self.estimated_cor,
                "research_width": self.research_width,
                "n_reconstruction": self.n_reconstruction,
                "mode": self.mode.value,
            }
        )
        return ddict

    def load_from_dict(self, dict_: dict):
        super().load_from_dict(dict_)
        if "research_width" in dict_:
            self.research_width = dict_["research_width"]
        if "estimated_cor" in dict_:
            self.estimated_cor = dict_["estimated_cor"]
        if "n_reconstruction" in dict_:
            self.n_reconstruction = dict_["n_reconstruction"]
        if "mode" in dict_:
            self.mode = ReconstructionMode(dict_["mode"])

    @staticmethod
    def from_dict(dict_):
        params = SAAxisParams()
        params.load_from_dict(dict_=dict_)
        return params

    def check_configuration(self):
        """
        Insure all requested information for processing the SAAXis are here.
        :raises: ValueError if some information are missing
        """
        missing_information = []
        if self.cors is None or len(self.cors) == 0:
            missing_information.append("no values for center of rotation provided")
        if self.slice_indexes is None:
            missing_information.append("slice index not provided")
        if len(missing_information) > 0:
            missing_information_str = " ; ".join(missing_information)
            raise ValueError(
                f"Some informations are missing: {missing_information_str}"
            )

    @staticmethod
    def compute_cors(estimated_cor, research_width, n_reconstruction):
        if estimated_cor is None:
            raise ValueError("No estimated cor provided")
        if estimated_cor == "middle":
            estimated_cor = 0
        if n_reconstruction % 2 == 0:
            n_reconstruction = n_reconstruction + 1
            # insure we have an odd number of cor to insure the estimated
            # one is reconstructed
        if n_reconstruction == 1:
            return (estimated_cor,)
        return numpy.linspace(
            start=estimated_cor - research_width / 2.0,
            stop=estimated_cor + research_width / 2.0,
            num=n_reconstruction,
        )



from __future__ import annotations

import os
import numpy
import pytest

from tomwer.core.volume.tiffvolume import TIFFVolume

from tomoscan.esrf.volume.tiffvolume import has_tifffile
from tomoscan.volumebase import VolumeBase, SliceTuple

from tomwer.core.volume.hdf5volume import HDF5Volume
from tomwer.core.volume.edfvolume import EDFVolume
from tomwer.tasks.visualization.volume_viewer import VolumeViewerTask


volume_constructors = [
    EDFVolume,
    HDF5Volume,
]
if has_tifffile:
    volume_constructors.append(TIFFVolume)


@pytest.mark.parametrize("volume_input_as", (VolumeBase, str))
@pytest.mark.parametrize("volume_constructor", volume_constructors)
def test_volume_viewer(tmp_path, volume_constructor, volume_input_as):
    volume_data = numpy.ones((10, 20, 30))
    volume_metadata = {
        "nabu_config": {
            "reconstruction": {
                "method": "FBP",
            },
            "phase": {
                "method": "paganin",
                "delta_beta": 110.0,
            },
        },
        "processing_options": {
            "reconstruction": {
                "voxel_size_cm": (0.2, 0.2, 0.2),
                "rotation_axis_position": 104,
                "enable_halftomo": True,
                "fbp_filter_type": "Hilbert",
                "sample_detector_dist": 0.4,
            },
            "take_log": {
                "log_min_clip": 1.0,
                "log_max_clip": 10.0,
            },
        },
    }
    if volume_constructor is HDF5Volume:
        volume = HDF5Volume(
            data=volume_data,
            metadata=volume_metadata,
            file_path=os.path.join(tmp_path, "output_file.h5"),
            data_path="my_volume",
        )
    else:
        vol_basename = "my_vol"
        volume = volume_constructor(
            data=volume_data,
            metadata=volume_metadata,
            folder=os.path.join(tmp_path, vol_basename),
            volume_basename=vol_basename,
        )
    volume.save()

    if volume_input_as is VolumeBase:
        input_volume = volume
    elif volume_input_as is str:
        input_volume = volume.get_identifier().to_str()
    else:
        raise TypeError(f"'volume_input_as' should be [{str}, {VolumeBase}]")

    task = VolumeViewerTask(
        inputs={
            "volume": input_volume,
            "load_volume": True,
        }
    )
    task.execute()

    slices = task.get_output_value("slices")
    assert len(slices) == 9
    for slice_ in slices:
        assert isinstance(slice_, SliceTuple)
    assert isinstance(task.get_output_value("volume_metadata"), dict)
    loaded_volume = task.get_output_value("loaded_volume")
    assert isinstance(loaded_volume.data, numpy.ndarray)



from silx.gui import qt
from tomwer.gui.control.observations import ScanObservation
from tomwer.tests.conftest import qtapp  # noqa F401
from tomwer.tests.conftest import nxtomo_scan_180  # noqa F401


def test_ScanObservation(qtapp, nxtomo_scan_180):  # noqa F811
    widget = ScanObservation(parent=None)
    widget.addObservation(
        scan=nxtomo_scan_180,
    )

    model = widget.observationTable.model()
    # check time stamp
    assert len(model.data(model.createIndex(0, 0), qt.Qt.DisplayRole)) == 8
    # check type
    assert model.data(model.createIndex(0, 1), qt.Qt.DisplayRole) == "hdf5"
    # check n projections
    assert model.data(model.createIndex(0, 2), qt.Qt.DisplayRole) == 0
    # check status
    assert model.data(model.createIndex(0, 3), qt.Qt.DisplayRole) == "starting"
    # check identifier
    assert "NXtomo" in model.data(model.createIndex(0, 4), qt.Qt.DisplayRole)



from __future__ import annotations

import os
from nabu.pipeline.config import parse_nabu_config_file
from tomwer.gui.configuration.action import (
    BasicConfigurationAction,
    ExpertConfigurationAction,
    MinimalisticConfigurationAction,
)
from silx.gui import qt
from nabu.pipeline.config import get_default_nabu_config
from nabu.pipeline.fullfield.nabu_config import (
    nabu_config as nabu_fullfield_default_config,
)
from tomwer.gui.reconstruction.nabu.slices import NabuWidget
from tomwer.gui.configuration.level import ConfigurationLevel


class NabuWidgetWithToolbar(qt.QMainWindow):
    sigConfigChanged = qt.Signal()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setWindowFlags(qt.Qt.Widget)

        self._nabuSettings = NabuWidget(parent=self)
        self.setCentralWidget(self._nabuSettings)

        self._createNabuSettingsToolbar()
        # connect signal / slot
        self._nabuSettings.sigConfigChanged.connect(self.sigConfigChanged)

    def _createNabuSettingsToolbar(self):
        style = qt.QApplication.style()

        # add toolbar
        toolbar = qt.QToolBar(self)
        toolbar.setMovable(False)
        toolbar.setFloatable(False)
        self.addToolBar(qt.Qt.TopToolBarArea, toolbar)

        # add configuration mode
        self.__configurationModesAction = qt.QAction(self)
        self.__configurationModesAction.setCheckable(False)
        menu = qt.QMenu(self)
        self.__configurationModesAction.setMenu(menu)
        toolbar.addAction(self.__configurationModesAction)

        self.__configurationModesGroup = qt.QActionGroup(self)
        self.__configurationModesGroup.setExclusive(True)
        self.__configurationModesGroup.triggered.connect(self._userModeChanged)

        self._minimalisticAction = MinimalisticConfigurationAction(toolbar)
        menu.addAction(self._minimalisticAction)
        self.__configurationModesGroup.addAction(self._minimalisticAction)
        self._basicConfigAction = BasicConfigurationAction(toolbar)
        menu.addAction(self._basicConfigAction)
        self.__configurationModesGroup.addAction(self._basicConfigAction)
        self._expertConfiguration = ExpertConfigurationAction(toolbar)
        menu.addAction(self._expertConfiguration)
        self.__configurationModesGroup.addAction(self._expertConfiguration)

        # reset configuration option
        self.__resetAction = qt.QAction(self)
        self.__resetAction.setToolTip("reset nabu configuration")
        reset_icon = style.standardIcon(qt.QStyle.SP_DialogResetButton)
        self.__resetAction.setIcon(reset_icon)
        toolbar.addAction(self.__resetAction)
        self.__resetAction.triggered.connect(self._resetParameters)

        # toolbar spacer
        self.__tSpacer = qt.QWidget(toolbar)
        self.__tSpacer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Preferred)
        toolbar.addWidget(self.__tSpacer)

        # connect signal / slot
        self._nabuSettings.sigConfigChanged.connect(self.sigConfigChanged)

        # set up
        self._resetParameters()
        self._basicConfigAction.setChecked(True)
        self._userModeChanged()

    def _resetParameters(self):
        # reset nabu settings
        default_config = get_default_nabu_config(nabu_fullfield_default_config)
        self._nabuSettings._configuration._preProcessingWidget._sinoRingsOpts.resetConfiguration()
        default_config["tomwer_slices"] = "middle"
        default_config["preproc"]["ccd_filter_enabled"] = False
        default_config["preproc"]["double_flatfield"] = False
        default_config["preproc"]["flatfield"] = True
        default_config["preproc"]["take_logarithm"] = True
        self._nabuSettings.setConfiguration(default_config)

    def _userModeChanged(self, *args, **kwargs):
        selectedAction = self.__configurationModesGroup.checkedAction()
        self.__configurationModesAction.setIcon(selectedAction.icon())
        self.__configurationModesAction.setToolTip(selectedAction.tooltip())
        self._nabuSettings.setConfigurationLevel(self.getConfigurationLevel())
        self.sigConfigChanged.emit()

    def _loadParameters(self):
        inputFile = self.askForNabuconfigFile(acceptMode=qt.QFileDialog.AcceptOpen)

        if inputFile and os.path.exists(inputFile):
            config = parse_nabu_config_file(inputFile)
            self._nabuSettings.setConfiguration(config)

    def setConfigurationLevel(self, level):
        level = ConfigurationLevel(level)
        if level == ConfigurationLevel.REQUIRED:
            self._minimalisticAction.setChecked(True)
        elif level == ConfigurationLevel.ADVANCED:
            self._expertConfiguration.setChecked(True)
        elif level == ConfigurationLevel.OPTIONAL:
            self._basicConfigAction.setChecked(True)
        else:
            raise ValueError("Level not recognize")
        self._userModeChanged()

    def getConfigurationLevel(self):
        if self._basicConfigAction.isChecked():
            return ConfigurationLevel.OPTIONAL
        elif self._expertConfiguration.isChecked():
            return ConfigurationLevel.ADVANCED
        elif self._minimalisticAction.isChecked():
            return ConfigurationLevel.REQUIRED
        else:
            raise ValueError("Level not recognize")

    # expose API
    def hideSlicesInterface(self):
        return self._nabuSettings.hideSlicesInterface()

    def hidePaganinInterface(self):
        return self._nabuSettings.hidePaganinInterface()

    def setConfiguration(self, *args, **kwargs):
        self._nabuSettings.setConfiguration(*args, **kwargs)

    def getConfiguration(self):
        return self._nabuSettings.getConfiguration()

    def getMode(self):
        return self._nabuSettings.getMode()

    def setMode(self, *args, **kwargs):
        self._nabuSettings.setMode(*args, **kwargs)

    def setScan(self, scan):
        self._nabuSettings.setScan(scan=scan)



"""
contains gui relative to semi-automatic axis calculation
"""

from __future__ import annotations

import pint
from typing import Iterable

from silx.gui import qt
from tomwer.gui.utils.qt_utils import block_signals

_ureg = pint.get_application_registry()


class DimensionWidget(qt.QGroupBox):
    """
    Simple widget to display value over 3 dimensions

    :param parent:
    :param title: QGroupBox title
    :param dims_name: name of the dimension. If set will be store in each
                      QDoubleLine prefix
    :param dims_colors: color associated to the three dimensions if any
    """

    valuesChanged = qt.Signal()
    """Signal emitted when a value change"""

    def __init__(
        self, parent=None, title=None, dims_name=None, dims_colors=None, title_size=10
    ):
        qt.QGroupBox.__init__(self, parent)
        self.setFont(qt.QFont("Arial", title_size))
        assert title is not None
        assert dims_name is None or (
            isinstance(dims_name, Iterable) and len(dims_name) == 3
        )
        assert dims_colors is None or (
            isinstance(dims_colors, Iterable) and len(dims_colors) == 3
        )
        self._displayUnit = _ureg.millimeter
        self._dim0Value: pint.Quantity = 1.0 * self._displayUnit
        self._dim1Value: pint.Quantity = 1.0 * self._displayUnit
        self._dim2Value: pint.Quantity = 1.0 * self._displayUnit
        # defined unit to display values. Always stored in m (International
        # System)
        ## set GUI
        self.setTitle(title)
        self.setLayout(qt.QHBoxLayout())
        # dim 0
        self._dim0ValueQLE = qt.QDoubleSpinBox(self)
        if dims_name is not None:
            self._dim0ValueQLE.setPrefix(dims_name[0])
        self._dim0ValueQLE.setRange(0, 999999999999)
        self._dim0ValueQLE.setDecimals(10)
        self._dim0ValueQLE.setSingleStep(0.0001)
        self._dim0ValueQLE.setValue(self._dim0Value.magnitude)
        if dims_colors is not None:
            stylesheet = f"background-color: {dims_colors[0]}"
            self._dim0ValueQLE.setStyleSheet(stylesheet)
        self.layout().addWidget(self._dim0ValueQLE)
        # dim 1
        self._dim1ValueQLE = qt.QDoubleSpinBox(self)
        if dims_name is not None:
            self._dim1ValueQLE.setPrefix(dims_name[1])
        self._dim1ValueQLE.setRange(0, 999999999999)
        self._dim1ValueQLE.setDecimals(10)
        self._dim1ValueQLE.setSingleStep(0.0001)
        self._dim1ValueQLE.setValue(self._dim1Value.magnitude)
        if dims_colors is not None:
            stylesheet = f"background-color: {dims_colors[1]}"
            self._dim1ValueQLE.setStyleSheet(stylesheet)
        self.layout().addWidget(self._dim1ValueQLE)
        # dim 2
        self._dim2ValueQLE = qt.QDoubleSpinBox(self)
        if dims_name is not None:
            self._dim2ValueQLE.setPrefix(dims_name[2])
        self._dim2ValueQLE.setRange(0, 999999999999)
        self._dim2ValueQLE.setDecimals(10)
        self._dim2ValueQLE.setSingleStep(0.0001)
        self._dim2ValueQLE.setValue(self._dim2Value.magnitude)
        if dims_colors is not None:
            stylesheet = f"background-color: {dims_colors[2]}"
            self._dim2ValueQLE.setStyleSheet(stylesheet)
        self.layout().addWidget(self._dim2ValueQLE)

        # set up
        self.setUnit(self._displayUnit)

        # connect signal / slot
        self._dim0ValueQLE.editingFinished.connect(self._userSetDim0)
        self._dim1ValueQLE.editingFinished.connect(self._userSetDim1)
        self._dim2ValueQLE.editingFinished.connect(self._userSetDim2)

    def setUnit(self, unit: pint.Unit):
        """
        define with which unit we should display the size
        :param unit: metric to be used for display. Internally this is always stored using the international metric system
        """
        if not isinstance(unit, pint.Unit):
            raise TypeError(
                f"unit is expected be an instance of {pint.Unit}. Got {type(unit)}"
            )
        for widget in (self._dim0ValueQLE, self._dim1ValueQLE, self._dim2ValueQLE):
            widget.setSuffix(f"{self._displayUnit:.4f~P}")
        # convert current values
        self._dim0Value = self._dim0Value.magnitude * self._displayUnit
        self._dim1Value = self._dim1Value.magnitude * self._displayUnit
        self._dim2Value = self._dim2Value.magnitude * self._displayUnit
        # update displayed values
        with block_signals(self):
            self._dim0ValueQLE.setValue(self._dim0Value.magnitude)
            self._dim1ValueQLE.setValue(self._dim1Value.magnitude)
            self._dim2ValueQLE.setValue(self._dim2Value.magnitude)

    def setQuantities(
        self,
        dim0: pint.Quantity,
        dim1: pint.Quantity,
        dim2: pint.Quantity,
    ) -> None:
        """

        :param dim0: value to dim0
        :param dim1: value to dim1
        :param dim2: value to dim2
        :param unit: unit used for the provided values
        """
        with block_signals(self):
            self.setDim0Quantity(value=dim0)
            self.setDim1value(value=dim1)
            self.setDim2Quantity(value=dim2)
        self.valuesChanged.emit()

    def getQuantities(
        self, cast_unit_to: None | pint.Unit = None
    ) -> tuple[pint.Quantity]:
        """
        :param cast_unit_to: if given will cast the quantities to the requested unit

        :return: (dim0 value, dim1 value, dim2 value)
        """
        if cast_unit_to is None:
            return (
                self.getDim0Quantity(),
                self.getDim1Quantity(),
                self.getDim2Quantity(),
            )
        else:
            return (
                self.getDim0Quantity().to(cast_unit_to),
                self.getDim1Quantity().to(cast_unit_to),
                self.getDim2Quantity().to(cast_unit_to),
            )

    def getDim0Quantity(self) -> tuple:
        """Return Dim 0 value and unit. Unit is always meter"""
        return self._dim0Value

    def setDim0Quantity(self, value: pint.Quantity):
        """

        :param value: value to set to dim 0.
        :return:
        """
        assert isinstance(value, pint.Quantity)
        self._dim0Value = value
        with block_signals(self):
            self._dim0ValueQLE.setValue(self._dim0Value.magnitude)
        self.valuesChanged.emit()

    def getDim1Quantity(self) -> pint.Quantity:
        """Return Dim 1 Quantity"""
        return self._dim1Value

    def setDim1value(self, value: pint.Quantity):
        """

        :param value: value to set to dim 1.
        :return:
        """
        assert isinstance(value, pint.Quantity)
        self._dim1Value = value
        with block_signals(self):
            self._dim1ValueQLE.setValue(self._dim1Value.magnitude)
        self.valuesChanged.emit()

    def getDim2Quantity(self) -> pint.Quantity:
        """Return Dim 2 value and unit. Unit is always meter"""
        return self._dim2Value

    def setDim2Quantity(self, value: pint.Quantity):
        """

        :param value: value to set to dim 2.
        :return:
        """
        assert isinstance(value, pint.Quantity)
        self._dim2Value = value
        with block_signals(self):
            self._dim2ValueQLE.setValue(self._dim2Value.magnitude)
        self.valuesChanged.emit()

    def _valuesChanged(self, *args, **kwargs):
        self.valuesChanged.emit()

    def _userSetDim0(self):
        """callback when the user modify the dim 0 QDSP"""
        with block_signals(self):
            self._dim0Value = self._dim0ValueQLE.value() * self._displayUnit
        self.valuesChanged.emit()

    def _userSetDim1(self):
        """callback when the user modify the dim 1 QDSP"""
        with block_signals(self):
            self._dim1Value = self._dim1ValueQLE.value() * self._displayUnit
        self.valuesChanged.emit()

    def _userSetDim2(self):
        """callback when the user modify the dim 2 QDSP"""
        with block_signals(self):
            self._dim2Value = self._dim2ValueQLE.value() * self._displayUnit
        self.valuesChanged.emit()



"""
project resources (icons... )
"""

from __future__ import annotations

import os
import sys
import contextlib
import atexit
from typing import NamedTuple
import importlib
import functools

# For packaging purpose, patch this variable to use an alternative directory
# E.g., replace with _RESOURCES_DIR = '/usr/share/silx/data'
_RESOURCES_DIR = None

# For packaging purpose, patch this variable to use an alternative directory
# E.g., replace with _RESOURCES_DIR = '/usr/share/silx/doc'
# Not in use, uncomment when functionality is needed
# _RESOURCES_DOC_DIR = None

# cx_Freeze frozen support
# See http://cx-freeze.readthedocs.io/en/latest/faq.html#using-data-files
if getattr(sys, "frozen", False):
    # Running in a frozen application:
    # We expect resources to be located either in a silx/resources/ dir
    # relative to the executable or within this package.
    _dir = os.path.join(os.path.dirname(sys.executable), "tomwer", "resources")
    if os.path.isdir(_dir):
        _RESOURCES_DIR = _dir


# Manage resource files life-cycle
_file_manager = contextlib.ExitStack()
atexit.register(_file_manager.close)


class _ResourceDirectory(NamedTuple):
    """Store a source of resources"""

    package_name: str
    forced_path: str | None = None


_TOMWER_DIRECTORY = _ResourceDirectory(
    package_name=__name__,
    forced_path=_RESOURCES_DIR,
)

_RESOURCE_DIRECTORIES = {}
_RESOURCE_DIRECTORIES["tomwer"] = _TOMWER_DIRECTORY


def _get_package_and_resource(
    resource, default_directory=None
) -> tuple[_ResourceDirectory, str]:
    """
    Return the resource directory class and a cleaned resource name without
    prefix.

    :param resource: Name of the resource with resource prefix.
    :param default_directory: If the resource is not prefixed, the resource
        will be searched on this default directory of the silx resource
        directory.
    :raises ValueError: If the resource name uses an unregistered resource
        directory name
    """
    if ":" in resource:
        prefix, resource = resource.split(":", 1)
    else:
        prefix = "tomwer"
        if default_directory is not None:
            resource = f"{default_directory}/{resource}"
    if prefix not in _RESOURCE_DIRECTORIES:
        raise ValueError("Resource '%s' uses an unregistred prefix", resource)
    resource_directory = _RESOURCE_DIRECTORIES[prefix]
    return resource_directory, resource


# Manage resource files life-cycle
_file_manager = contextlib.ExitStack()
atexit.register(_file_manager.close)


@functools.lru_cache(maxsize=None)
def _get_resource_filename(package: str, resource: str) -> str:
    """Returns path to requested resource in package

    :param package: Name of the package in which to look for the resource
    :param resource: Resource path relative to package using '/' path separator
    :return: Abolute resource path in the file system
    """
    # Caching prevents extracting the resource twice
    file_context = importlib.resources.as_file(
        importlib.resources.files(package) / resource
    )
    path = _file_manager.enter_context(file_context)
    return str(path.absolute())


def _resource_filename(resource: str, default_directory: str | None = None) -> str:
    """Return filename corresponding to resource.

    The existence of the resource is not checked.

    The resource name can be prefixed by the name of a resource directory. For
    example "silx:foo.png" identify the resource "foo.png" from the resource
    directory "silx". See also :func:`register_resource_directory`.

    :param resource: Resource path relative to resource directory
                     using '/' path separator. It can be either a file or
                     a directory.
    :param default_directory: If the resource is not prefixed, the resource
        will be searched on this default directory of the silx resource
        directory. It should only be used internally by silx.
    :return: Absolute resource path in the file system
    """
    resource_directory, resource_name = _get_package_and_resource(
        resource, default_directory=default_directory
    )

    if resource_directory.forced_path is not None:
        # if set, use this directory
        base_dir = resource_directory.forced_path
        resource_path = os.path.join(base_dir, *resource_name.split("/"))
        return resource_path

    return _get_resource_filename(resource_directory.package_name, resource_name)



from __future__ import annotations

from orangewidget import gui, settings, widget
from orangewidget.widget import Output
from silx.gui import qt
import os

from orangecontrib.tomwer.widgets.utils import WidgetLongProcessing
from tomwer.core.process.control.datalistener import DataListener
from tomwer.core.process.control.nxtomomill import H5ToNxProcess
from tomwer.core.scan.blissscan import BlissScan
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.gui.control.datalistener import DataListenerWidget
from tomwer.gui.utils.qt_utils import block_signals
from tomwer.synctools.stacks.control.datalistener import DataListenerProcessStack
from tomwer.utils import docstring

from nxtomomill.models.h52nx import H52nxModel

import logging
import signal
from typing import Iterable

logger = logging.getLogger(__name__)


class DataListenerBaseClass(
    widget.OWBaseWidget,
    WidgetLongProcessing,
    DataListener,
    openclass=True,
):
    """
    This widget is a mix-in class for widgets listening a scan.
    Scan decovery can come from `bliss-tomo tomoSync<https://gitlab.esrf.fr/tomo/bliss-tomo/-/blob/main/tomo_listener/TomoSynch.py>`
    or from `bliss data <https://gitlab.esrf.fr/bliss/blissdata>`
    """

    want_main_area = True
    want_control_area = False
    resizing_enabled = True

    _blissConfiguration = settings.Setting(dict())
    # to keep backward compatibility

    _nxtomo_cfg_file = settings.Setting(str())
    # to keep backward compatibility

    _static_input = settings.Setting(dict())

    class Outputs:
        data = Output(name="data", type=TomwerScanBase, doc="one scan to be process")

    def __init__(self, host_discovery: str | None, parent=None, uses_rpc: bool = False):
        widget.OWBaseWidget.__init__(self, parent)
        WidgetLongProcessing.__init__(self)
        DataListener.__init__(self)
        self._processingStack = DataListenerProcessStack()
        self._processingStack.sigComputationEnded.connect(self._signal_scan_ready)
        self._widget = DataListenerWidget(parent=self, host_discovery=host_discovery)
        self._mock = False
        self._is_using_rpc = uses_rpc

        self._box = gui.vBox(self.mainArea, self.name)
        layout = self._box.layout()
        layout.addWidget(self._widget)

        # signal / slot connection
        self._widget.sigActivate.connect(self._activated)
        self._widget.sigDeactivate.connect(self._deactivated)
        self._widget.sigConfigurationChanged.connect(self._jsonRPCConfigChanged)
        self._widget.sigCFGFileChanged.connect(self._nxtomoFileChanged)
        self._widget.sigAcquisitionEnded.connect(self._process_bliss_file_frm_tuple)
        self._widget.sigServerStopped.connect(self._serverStopped)
        self._widget.sigMechanicalFlipsChanged.connect(self._mechanicalFlipHaveChanged)

        # set up
        self._loadSettings()

        # for convenience start the listener when create it.
        # ONLY if 'BEACON_HOST' is defined or using json-rpc
        if self._is_using_rpc or bool(os.environ.get("BEACON_HOST", "")):
            self.activate(True, is_using_rpc=self._is_using_rpc)

    def _loadSettings(self):
        if "bliss_server_configuration" in self._static_input:  # pylint: disable=E1135
            bliss_configuration = self._static_input[  # pylint: disable=E1136
                "bliss_server_configuration"
            ]
        else:
            bliss_configuration = self._blissConfiguration
        if bliss_configuration != {}:
            self._widget.setBlissServerConfiguration(bliss_configuration)
        if "nxtomomill_cfg_file" in self._static_input:  # pylint: disable=E1135
            nxtomo_cfg_file = self._static_input[  # pylint: disable=E1136
                "nxtomomill_cfg_file"
            ]
        else:
            nxtomo_cfg_file = self._nxtomo_cfg_file
        self._widget.setCFGFilePath(nxtomo_cfg_file)
        if "output_dir" in self._static_input:  # pylint: disable=E1135
            self._widget.setOutputFolder(
                self._static_input["output_dir"]  # pylint: disable=E1136
            )
        mechanical_flips = self._static_input.get("mechanical_flips")
        if mechanical_flips is not None:  # pylint: disable=E1135
            lr_flip, ud_flip = mechanical_flips
            with block_signals(self._widget):
                self._widget.setMechanicalFlips(
                    left_right_flip=lr_flip, up_down_flip=ud_flip
                )

    def getNXTomomillConfiguration(self):
        cfg_file = self._widget.getCFGFilePath()

        def create_default_config():
            configuration = H52nxModel()
            configuration.bam_single_file = True
            configuration.no_master_file = True
            return configuration

        if cfg_file in (None, ""):
            config = create_default_config()
        else:
            try:
                config = H52nxModel.from_cfg_file(cfg_file)
            except Exception as e:
                logger.warning(f"Fail to load configuration file. Error is {e}")
                config = create_default_config()
        # apply mechanical flip
        config.mechanical_lr_flip, config.mechanical_ud_flip = (
            self._widget.getMechanicalFlips()
        )
        return config

    def _process_bliss_file_frm_tuple(self, t):
        master_file, entry, proposal_file, saving_file, success = t
        bliss_scan = BlissScan(
            master_file=master_file,
            entry=str(entry) + ".1",
            proposal_file=proposal_file,
            saving_file=saving_file,
        )
        configuration = self.getNXTomomillConfiguration()
        # overwrite output file
        configuration.output_file = H5ToNxProcess.deduce_output_file_path(
            bliss_scan.master_file,
            output_dir=self._widget.getOutputFolder(),
            scan=bliss_scan,
        )
        if success:
            self._processingStack.add(data=bliss_scan, configuration=configuration)
        else:
            pass

    def activate(self, activate=True, is_using_rpc: bool = False):

        with block_signals(self._widget):
            self.set_configuration(self._widget.getBlissServerConfiguration())
            self._widget.activate(activate=activate)
            super().activate(activate=activate, is_using_rpc=is_using_rpc)
            self.processing_state(activate, info="listener active")

    def _activated(self):
        self.activate(True, is_using_rpc=self._is_using_rpc)

    def _deactivated(self):
        self.activate(False, is_using_rpc=self._is_using_rpc)

    def _mechanicalFlipHaveChanged(self):
        self._static_input["mechanical_flips"] = self._widget.getMechanicalFlips()

    def _serverStopped(self):
        """
        Callback when the server is stopped
        """
        self.activate(False, is_using_rpc=self._is_using_rpc)

    def _signal_scan_ready(self, scan, future_tomo_obj):
        if scan is None:
            return
        assert isinstance(scan, Iterable)
        for s in scan:
            assert isinstance(s, TomwerScanBase)
            self.Outputs.data.send(s)

    def _ask_user_for_overwritting(self, file_path):
        msg = qt.QMessageBox(self)
        msg.setIcon(qt.QMessageBox.Question)
        types = qt.QMessageBox.Ok | qt.QMessageBox.Cancel
        msg.setStandardButtons(types)

        text = "NXtomomill will overwrite: \n %s. Do you agree ?" % file_path
        msg.setText(text)
        return msg.exec() == qt.QMessageBox.Ok

    def _jsonRPCConfigChanged(self):
        self._blissConfiguration = self._widget.getBlissServerConfiguration()
        self._static_input["bliss_server_configuration"] = (  # pylint: disable=E1137
            self._widget.getBlissServerConfiguration()
        )
        self._static_input["output_dir"] = (  # pylint: disable=E1137
            self._widget.getOutputFolder()
        )
        if self.is_active():
            self.activate(False, is_using_rpc=True)
            self.activate(True, is_using_rpc=True)

    def _nxtomoFileChanged(self, cfg_file):
        self._nxtomo_cfg_file = cfg_file
        self._static_input["nxtomomill_cfg_file"] = cfg_file  # pylint: disable=E1137

    def setMock(self, mock, acquisitions):
        self._mock = mock
        self._mock_acquisitions = acquisitions

    @docstring(DataListenerWidget.getHost)
    def getHost(self):
        return self._widget.getHost()

    @docstring(DataListenerWidget.getPort)
    def getPort(self):
        return self._widget.getPort()

    def close(self):
        signal.signal(signal.SIGINT, signal.SIG_DFL)
        self.activate(False)
        super().close()

    def _get_n_scan_observe(self):
        return self._widget._observationWidget.observationTable.model().rowCount()

    def _get_n_scan_finished(self):
        return self._widget._historyWindow.scanHistory.model().rowCount()



"""
(Ewoks)Orange Widgets for tomography not matching any other category
"""

NAME = "other"

ID = "orangecontrib.tomwer.widgets.other"

DESCRIPTION = "some utils widgets"

LONG_DESCRIPTION = "some utils widgets"

ICON = "../../widgets/icons/tomwer_other.png"

BACKGROUND = "#C0CCFF"

PRIORITY = 20



# coding: utf-8
from __future__ import annotations


import os
import threading
import uuid

from orangewidget import gui, widget
from orangewidget.settings import Setting
from orangewidget.widget import Output
from silx.gui import qt
from silx.gui.utils import concurrent

from tomwer.core.scan.blissscan import BlissScan
from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.utils.scanutils import MockBlissAcquisition, MockEDF, MockNXtomo
from tomwer.gui.debugtools.datasetgenerator import DatasetGeneratorDialog


class DatasetGeneratorOW(widget.OWBaseWidget, openclass=True):
    """
    A simple widget to generate on the fly and 'continuously' datasets
    """

    name = "random data generator"
    id = "orangecontrib.tomwer.widgets.debugtools.datasetgeneratorow"
    description = "create on the fly dataset"
    icon = "icons/hammer.png"
    priority = 250
    keywords = ["tomography", "file", "tomwer", "dataset", "debug"]

    want_main_area = True
    want_control_area = False
    resizing_enabled = True

    _settings = Setting(dict())

    class Outputs:
        data = Output(name="data", type=object)

    def __init__(self, parent=None):
        widget.OWBaseWidget.__init__(self, parent=parent)
        self._layout = gui.vBox(self.mainArea, self.name).layout()
        self.generatorGUI = DatasetGeneratorDialog(parent=self)
        self._layout.addWidget(self.generatorGUI)
        self._displayTimer = qt.QTimer()

        self._loadSettings()

        # connect signal / slot
        self.generatorGUI.sigGenerationStopped.connect(self._stop)
        self.generatorGUI.sigGenerationStarted.connect(self._start)
        self.generatorGUI.sigConfigChanged.connect(self._reset)
        self.generatorGUI.sigCreateOne.connect(self._generate)
        #  connection with the interface to update settings
        self.generatorGUI.sigConfigChanged.connect(self._saveSettings)

    def _start(self):
        self._displayTimer.timeout.connect(self._generateAndRestart)
        self._displayTimer.start(self.generatorGUI.getTimeout())

    def _stop(self):
        self._displayTimer.timeout.disconnect(self._generateAndRestart)
        self._displayTimer.stop()

    def _reset(self):
        if self._displayTimer.isActive():
            self._stop()
            self._start()

    def _generateAndRestart(self):
        self._generate()
        self._displayTimer.start(self.generatorGUI.getTimeout())

    def _generate(self):
        scan_path = os.path.join(
            self.generatorGUI.getRootDir(),
            str(uuid.uuid4()).split("-")[0],
        )
        thread = _DatasetGeneratorThread()
        thread.init(
            type_to_generate=self.generatorGUI.getTypeToGenerate(),
            scan_path=scan_path,
            callback=self.Outputs.data.send,
            n_proj=self.generatorGUI.getNProj(),
            create_ini_dark=self.generatorGUI.isDarkNeededAtBeginning(),
            create_ini_flat=self.generatorGUI.isFlatNeededAtBeginning(),
            dims=self.generatorGUI.getFrameDims(),
        )
        thread.start()
        return thread

    def _loadSettings(self):
        self.generatorGUI.setConfiguration(self._settings)

    def _saveSettings(self):
        self._settings = self.generatorGUI.getConfiguration()


class _DatasetGeneratorThread(threading.Thread):
    def init(
        self,
        type_to_generate,
        scan_path,
        callback,
        n_proj,
        create_ini_dark,
        create_ini_flat,
        dims,
    ):
        self.type_to_generate = type_to_generate
        self.callback = callback
        self.scan_path = scan_path
        self.n_proj = n_proj
        self.create_ini_dark = create_ini_dark
        self.create_ini_flat = create_ini_flat
        self.dims = dims

    def run(self):
        if self.type_to_generate == EDFTomoScan.__name__:
            scan = MockEDF.mockScan(
                scanID=self.scan_path, nRadio=self.n_proj, dim=self.dims[0]
            )
        elif self.type_to_generate == NXtomoScan.__name__:
            scan = MockNXtomo(
                scan_path=self.scan_path,
                n_proj=self.n_proj,
                n_ini_proj=self.n_proj,
                create_ini_dark=self.create_ini_dark,
                create_ini_flat=self.create_ini_flat,
                dim=self.dims[0],
            ).scan
        elif self.type_to_generate == BlissScan.__name__:
            n_darks = 10 if self.create_ini_dark else 0
            n_flats = 10 if self.create_ini_flat else 0
            mock = MockBlissAcquisition(
                n_sample=1,
                n_sequence=1,
                n_scan_per_sequence=1,
                n_darks=n_darks,
                n_flats=n_flats,
                output_dir=self.scan_path,
            )
            scan = mock.create_bliss_scan()
        else:
            raise ValueError(f"type to generate {self.type_to_generate} not recognized")
        concurrent.submitToQtMainThread(self.callback, scan)
